from dataclasses import dataclass
import ipaddress
from enum import Enum
from typing import Optional, List, Tuple, Dict
from oonidata.experiments.control import DNSControl, HTTPControl, TCPControl
from oonidata.experiments.experiment_result import (
    BlockingEvent,
    BlockingType,
    ExperimentResult,
    fp_scope_to_outcome,
    make_base_result_meta,
)

from oonidata.fingerprintdb import FingerprintDB
from oonidata.netinfo import NetinfoDB


from oonidata.observations import (
    DNSObservation,
    HTTPObservation,
    NettestObservation,
    TCPObservation,
    TLSObservation,
)

import logging

log = logging.getLogger("oonidata.processing")


def is_dns_consistent(
    dns_o: DNSObservation, dns_ctrl: DNSControl, netinfodb: NetinfoDB
) -> Tuple[bool, float]:
    if not dns_o.answer:
        return False, 0

    try:
        ipaddress.ip_address(dns_o.answer)
    except ValueError:
        # Not an IP, er can't do much to validate it
        return False, 0

    if dns_o.answer in dns_ctrl.tls_consistent_answers:
        return True, 1.0

    baseline_asns = set()
    baseline_as_org_names = set()

    for ip in dns_ctrl.tls_consistent_answers:
        ip_info = netinfodb.lookup_ip(dns_o.timestamp, ip)
        if ip_info:
            baseline_asns.add(ip_info.as_info.asn)
            baseline_as_org_names.add(ip_info.as_info.as_org_name.lower())

    if dns_o.answer_asn in baseline_asns:
        return True, 0.9

    # XXX maybe with the org_name we can also do something like levenshtein
    # distance to get more similarities
    if (
        dns_o.answer_as_org_name
        and dns_o.answer_as_org_name.lower() in baseline_as_org_names
    ):
        return True, 0.9

    other_answers = dns_ctrl.answers_map.copy()
    other_answers.pop(dns_o.probe_cc, None)
    other_ips = {}
    other_asns = {}
    for answer_list in other_answers.values():
        for _, ip in answer_list:

            other_ips[ip] = other_ips.get(ip, 0)
            other_ips[ip] += 1
            if ip is None:
                log.error(f"Missing ip for {dns_o.domain_name}")
                continue
            ip_info = netinfodb.lookup_ip(dns_o.timestamp, ip)
            if ip_info:
                asn = ip_info.as_info.asn
                other_asns[asn] = other_asns.get(ip, 0)
                other_asns[asn] += 1

    if dns_o.answer in other_ips:
        x = other_ips[dns_o.answer]
        # This function was derived by looking for an exponential function in
        # the form f(x) = c1*a^x + c2 and solving for f(0) = 0 and f(10) = 1,
        # giving us a function in the form f(x) = (a^x - 1) / (a^10 - 1). We
        # then choose the magic value of 0.6 by looking for a solution in a
        # where f(1) ~= 0.5, doing a bit of plots and choosing a curve that
        # looks reasonably sloped.
        y = (pow(0.5, x) - 1) / (pow(0.5, 10) - 1)
        return True, min(0.9, 0.8 * y)

    if dns_o.answer in other_asns:
        x = other_asns[dns_o.answer_asn]
        y = (pow(0.5, x) - 1) / (pow(0.5, 10) - 1)
        return True, min(0.8, 0.7 * y)

    x = len(baseline_asns)
    y = (pow(0.5, x) - 1) / (pow(0.5, 10) - 1)
    return False, min(0.9, 0.8 * y)


def make_website_tcp_blocking_event(
    tcp_o: TCPObservation, tcp_b: TCPControl
) -> BlockingEvent:
    blocking_type = BlockingType.OK
    blocking_detail = "ok"
    blocking_subject = f"{tcp_o.ip}:{tcp_o.port}"
    blocking_meta = {}
    confidence = 1.0

    if tcp_o.failure:
        unreachable_cc_asn = list(tcp_b.unreachable_cc_asn)
        try:
            unreachable_cc_asn.remove((tcp_o.probe_cc, tcp_o.probe_asn))
        except ValueError:
            log.info(
                "missing failure in tcp baseline. You are probably using a control derived baseline."
            )

        reachable_count = len(tcp_b.reachable_cc_asn)
        unreachable_count = len(unreachable_cc_asn)
        blocking_meta = {
            "unreachable_count": unreachable_count,
            "reachable_count": reachable_count,
        }
        if reachable_count > unreachable_count:
            # We are adding back 1 because we removed it above and it avoid a divide by zero
            confidence = reachable_count / (reachable_count + unreachable_count + 1)
            blocking_type = BlockingType.BLOCKED
        elif unreachable_count > reachable_count:
            confidence = (unreachable_count + 1) / (
                reachable_count + unreachable_count + 1
            )
            blocking_type = BlockingType.BLOCKED

        # TODO: should we bump up the confidence in the case of connection_reset?
        blocking_detail = f"tcp.{tcp_o.failure}"

    return BlockingEvent(
        blocking_type=blocking_type,
        blocking_subject=blocking_subject,
        blocking_detail=blocking_detail,
        blocking_meta=blocking_meta,
        confidence=confidence,
    )


def make_website_dns_blocking_event(
    dns_o: DNSObservation,
    dns_ctrl: DNSControl,
    fingerprintdb: FingerprintDB,
    netinfodb: NetinfoDB,
) -> BlockingEvent:

    blocking_subject = dns_o.domain_name

    if dns_o.fingerprint_id:
        fp = fingerprintdb.get_fp(dns_o.fingerprint_id)
        blocking_type = fp_scope_to_outcome(fp.scope)
        confidence = 1.0
        # If we see the fingerprint in an unexpected country we should
        # significantly reduce the confidence in the block
        if (
            dns_o.probe_cc
            and fp.expected_countries
            and len(fp.expected_countries) > 0
            and dns_o.probe_cc not in fp.expected_countries
        ):
            log.debug(
                f"Inconsistent probe_cc vs expected_countries {dns_o.probe_cc} != {fp.expected_countries}"
            )
            confidence = 0.7

        return BlockingEvent(
            blocking_type=blocking_type,
            blocking_subject=blocking_subject,
            blocking_detail="dns.inconsistent.blockpage",
            blocking_meta={"ip": dns_o.answer},
            confidence=confidence,
        )

    elif dns_o.answer_is_bogon and len(dns_ctrl.tls_consistent_answers) > 0:
        return BlockingEvent(
            blocking_type=BlockingType.BLOCKED,
            blocking_subject=blocking_subject,
            blocking_detail="dns.inconsistent.bogon",
            blocking_meta={"ip": dns_o.answer},
            confidence=0.9,
        )

    elif dns_o.failure:
        failure_cc_asn = list(dns_ctrl.failure_cc_asn)
        try:
            failure_cc_asn.remove((dns_o.probe_cc, dns_o.probe_asn))
        except ValueError:
            log.info(
                "missing failure for the probe in the baseline. You are probably using a control derived baseline."
            )

        failure_count = len(failure_cc_asn)
        ok_count = len(dns_ctrl.ok_cc_asn)

        if dns_o.failure == "dns_nxdomain_error":
            nxdomain_cc_asn = list(dns_ctrl.nxdomain_cc_asn)
            try:
                nxdomain_cc_asn.remove((dns_o.probe_cc, dns_o.probe_asn))
            except ValueError:
                log.info(
                    "missing nx_domain failure for the probe in the baseline. You are probably using a control derived baseline."
                )

            nxdomain_count = len(nxdomain_cc_asn)
            if ok_count > nxdomain_count:
                # We give a bit extra weight to an NXDOMAIN compared to other failures
                confidence = ok_count / (ok_count + nxdomain_count + 1)
                confidence = min(0.8, confidence * 1.5)
                blocking_type = BlockingType.BLOCKED
                blocking_detail = "dns.inconsistent.nxdomain"
            else:
                confidence = (nxdomain_count + 1) / (ok_count + nxdomain_count + 1)
                blocking_type = BlockingType.DOWN
                blocking_detail = "dns.inconsistent.nxdomain"
        elif ok_count > failure_count:
            confidence = ok_count / (ok_count + failure_count + 1)
            blocking_type = BlockingType.BLOCKED
            blocking_detail = f"dns.{dns_o.failure}"
        else:
            confidence = (failure_count + 1) / (ok_count + failure_count + 1)
            blocking_type = BlockingType.DOWN
            blocking_detail = f"dns.{dns_o.failure}"

        return BlockingEvent(
            blocking_type=blocking_type,
            blocking_subject=blocking_subject,
            blocking_detail=blocking_detail,
            blocking_meta={"ip": dns_o.answer},
            confidence=0.9,
        )

    elif dns_o.is_tls_consistent == False:
        return BlockingEvent(
            blocking_type=BlockingType.BLOCKED,
            blocking_subject=blocking_subject,
            blocking_detail="dns.inconsistent.tls_mismatch",
            blocking_meta={"ip": dns_o.answer, "why": "tls_inconsistent"},
            confidence=0.9,
        )

    elif dns_o.is_tls_consistent == None:
        # If we are in this case, it means we weren't able to determine the
        # consistency of the DNS query using TLS. This is the case either
        # because the tested site is not in HTTPS and therefore we didn't
        # generate a TLS measurement for it or because the target IP isn't
        # listening on HTTPS (which is quite fishy).
        # In either case we should flag these with being somewhat likely to be
        # blocked.
        ip_based_consistency, consistency_confidence = is_dns_consistent(
            dns_o, dns_ctrl, netinfodb
        )
        if ip_based_consistency is False and consistency_confidence > 0:
            confidence = consistency_confidence
            blocking_detail = "dns.inconsistent.generic"
            # If the answer ASN is the same as the probe_asn, it's more likely
            # to be a blockpage
            if dns_o.answer_asn == dns_o.probe_asn:
                blocking_detail = "dns.inconsistent.asn_match"
                confidence = 0.8
            # same for the answer_cc
            elif dns_o.answer_as_cc == dns_o.probe_cc:
                blocking_detail = "dns.inconsistent.cc_match"
                confidence = 0.7

            return BlockingEvent(
                blocking_type=BlockingType.BLOCKED,
                blocking_subject=blocking_subject,
                blocking_detail=blocking_detail,
                blocking_meta={"ip": dns_o.answer, "why": "tls_inconsistent"},
                confidence=0.9,
            )

    # No blocking detected
    return BlockingEvent(
        blocking_type=BlockingType.OK,
        blocking_subject=blocking_subject,
        blocking_detail="ok",
        blocking_meta={"ip": dns_o.answer},
        confidence=0.8,
    )


def make_website_tls_blocking_event(
    tls_o: TLSObservation, prev_be: List[BlockingEvent]
) -> Optional[BlockingEvent]:
    blocking_subject = tls_o.domain_name

    if tls_o.is_certificate_valid == False:
        # We only consider it to be a TLS level verdict in cases when there is a
        # certificate mismatch, but there was no DNS inconsistency.
        # If the DNS was inconsistent, we will just count the DNS verdict
        if (
            len(
                list(
                    filter(
                        lambda v: v.blocking_detail.startswith("dns.")
                        and v.blocking_meta.get("ip") == tls_o.ip,
                        prev_be,
                    )
                )
            )
            > 0
        ):
            return

        # TODO: this is wrong. We need to consider the baseline to establish TLS
        # MITM, because the cert might be invalid also from other location (eg.
        # it expired) and not due to censorship.
        return BlockingEvent(
            blocking_type=BlockingType.BLOCKED,
            blocking_subject=blocking_subject,
            blocking_detail="tls.mitm",
            blocking_meta={"why": "invalid certificate"},
            confidence=0.8,
        )

    elif tls_o.failure:
        if (
            len(
                list(
                    filter(
                        lambda v: v.blocking_detail.startswith("tcp.")
                        and v.blocking_subject == f"{tls_o.ip}:443",
                        prev_be,
                    )
                )
            )
            > 0
        ):
            return

        # We only consider it to be a TLS level verdict if we haven't seen any
        # blocks in TCP
        blocking_detail = f"tls.{tls_o.failure}"
        confidence = 0.5

        if tls_o.tls_handshake_read_count == 0 and tls_o.tls_handshake_write_count == 1:
            # This means we just wrote the TLS ClientHello, let's give it a bit
            # more confidence in it being a block
            confidence = 0.7

        if tls_o.failure in ("connection_closed", "connection_reset"):
            confidence += 0.2

        return BlockingEvent(
            blocking_type=BlockingType.BLOCKED,
            blocking_subject=blocking_subject,
            blocking_detail=blocking_detail,
            blocking_meta={},
            confidence=confidence,
        )

    return BlockingEvent(
        blocking_type=BlockingType.OK,
        blocking_subject=blocking_subject,
        blocking_detail="ok",
        blocking_meta={},
        confidence=1.0,
    )


def make_website_http_blocking_event(
    http_o: HTTPObservation,
    http_ctrl: HTTPControl,
    prev_be: List[BlockingEvent],
    fingerprintdb: FingerprintDB,
) -> Optional[BlockingEvent]:
    blocking_subject = http_o.request_url

    ok_be = BlockingEvent(
        blocking_type=BlockingType.OK,
        blocking_subject=blocking_subject,
        blocking_detail="ok",
        blocking_meta={},
        confidence=0.8,
    )

    detail_prefix = "http."
    if http_o.request_is_encrypted:
        detail_prefix = "https."

    if http_o.failure:
        # For HTTP requests we ignore cases in which we detected the blocking
        # already to be happening via DNS or TCP.
        if not http_o.request_is_encrypted and (
            len(
                list(
                    filter(
                        lambda v: v.blocking_detail.startswith("dns.")
                        or (
                            v.blocking_detail.startswith("tcp.")
                            and v.blocking_subject.endswith(":80")
                        ),
                        prev_be,
                    )
                )
            )
            > 0
        ):
            return

        # Similarly for HTTPS we ignore cases when the block is done via TLS or TCP
        if http_o.request_is_encrypted and (
            len(
                list(
                    filter(
                        lambda v: v.blocking_detail.startswith("dns.")
                        or (
                            v.blocking_detail.startswith("tcp.")
                            and v.blocking_subject.endswith(":443")
                        )
                        or v.blocking_detail.startswith("tls."),
                        prev_be,
                    )
                )
            )
            > 0
        ):
            return

        failure_cc_asn = list(http_ctrl.failure_cc_asn)
        try:
            failure_cc_asn.remove((http_o.probe_cc, http_o.probe_asn))
        except ValueError:
            log.info(
                "missing failure in http baseline. Either something is wrong or you are using a control derived baseline"
            )

        failure_count = len(failure_cc_asn)
        ok_count = len(http_ctrl.ok_cc_asn)
        if ok_count > failure_count:
            # We are adding back 1 because we removed it above and it avoid a divide by zero
            confidence = ok_count / (ok_count + failure_count + 1)
            blocking_type = BlockingType.BLOCKED
        else:
            confidence = (failure_count + 1) / (ok_count + failure_count + 1)
            blocking_type = BlockingType.DOWN
        blocking_detail = f"{detail_prefix}{http_o.failure}"

        return BlockingEvent(
            blocking_type=blocking_type,
            blocking_subject=blocking_subject,
            blocking_detail=blocking_detail,
            blocking_meta={},
            confidence=confidence,
        )

    elif http_o.response_matches_blockpage:
        blocking_type = BlockingType.BLOCKED
        blocking_meta = {}
        confidence = 0.7
        if http_o.request_is_encrypted:
            # Likely some form of server-side blocking
            blocking_type = BlockingType.SERVER_SIDE_BLOCK
            confidence = 0.5
        elif http_o.fingerprint_country_consistent:
            confidence = 1

        for fp_name in http_o.response_fingerprints:
            fp = fingerprintdb.get_fp(fp_name)
            if fp.scope:
                blocking_type = fp_scope_to_outcome(fp.scope)
                blocking_meta = {"fp_name": fp.name, "why": "matched fingerprint"}
                break

        blocking_detail = f"{detail_prefix}diff"
        return BlockingEvent(
            blocking_type=blocking_type,
            blocking_subject=blocking_subject,
            blocking_detail=blocking_detail,
            blocking_meta=blocking_meta,
            confidence=confidence,
        )

    elif not http_o.request_is_encrypted:
        if http_o.response_matches_false_positive:
            return ok_be
        if http_o.response_body_title == http_ctrl.response_body_title:
            return ok_be
        if http_o.response_body_meta_title == http_ctrl.response_body_meta_title:
            return ok_be
        if http_o.response_body_sha1 == http_ctrl.response_body_sha1:
            return ok_be

        if (
            http_o.response_body_length
            and http_ctrl.response_body_length
            and (
                (http_o.response_body_length + 1.0)
                / (http_ctrl.response_body_length + 1.0)
                < 0.7
            )
        ):
            blocking_detail = f"{detail_prefix}diff.body"
            return BlockingEvent(
                blocking_type=BlockingType.BLOCKED,
                blocking_subject=blocking_subject,
                blocking_detail=blocking_detail,
                blocking_meta={"why": "inconsistent body length"},
                confidence=0.6,
            )

    return ok_be


@dataclass
class WebsiteExperimentResult(ExperimentResult):
    domain_name: str
    website_name: str


def make_website_experiment_result(
    nt_o: NettestObservation,
    dns_o_list: List[DNSObservation],
    dns_ctrl: DNSControl,
    fingerprintdb: FingerprintDB,
    netinfodb: NetinfoDB,
    tcp_o_list: List[TCPObservation],
    tcp_ctrl_map: Dict[str, TCPControl],
    tls_o_list: List[TLSObservation],
    http_o_list: List[HTTPObservation],
    http_ctrl_map: Dict[str, HTTPControl],
) -> WebsiteExperimentResult:
    """
    make_website_verdicts will yield many verdicts given some observations
    related to a website measurement.

    We MUST pass in DNS observations, but observations of other types are
    optional. This is to workaround the fact that not every version of OONI
    Probe was generating all types of observations.

    The order in which we compute the verdicts is important, since the knowledge
    of some form of blocking is relevant to being able to determine future
    methods of blocking.
    Examples of this include:
    * If you know that DNS is consistent and you see a TLS certificate
    validation error, you can conclude that it's a MITM
    * If you see that TCP connect is failing, you will not attribute a failing
    TLS to a TLS level interference (ex. SNI filtering)

    For some lists of observations we also need to pass in a baseline. The
    baseline is some groundtruth related to the targets being measured that are
    needed in order to draw some meaningful conclusion about it's blocking.
    We need this so that we are able to exclude instances in which the target is
    unavailable due to transient network failures.

    It is the job of who calls the make_website_verdicts function to build this
    baseline by either running queries against the database of observations or
    using some set of observations that are already in memory.

    The basic idea is that since we are going to be generating verdicts in
    windows of 24h, we would have to do the baselining only once for the 24h
    time window given a certain domain.
    """
    blocking_events = []
    observation_ids = []

    domain_name = dns_o_list[0].domain_name
    dns_blocking_events = []
    for dns_o in dns_o_list:
        observation_ids.append(dns_o.observation_id)

        assert (
            domain_name == dns_o.domain_name
        ), f"Inconsistent domain_name in dns_o {dns_o.domain_name}"
        dns_be = make_website_dns_blocking_event(
            dns_o, dns_ctrl, fingerprintdb, netinfodb
        )
        if dns_be:
            dns_blocking_events.append(dns_be)
        else:
            # If we didn't get a DNS blocking event from an observation, it means that
            # observation was a sign of everything being OK, hence we should
            # ignore all the previous DNS verdicts as likely false positives and
            # just consider no DNS level censorship to be happening.
            # TODO: probably we want to just reduce the confidence of the DNS
            # level blocks in this case by some factor.
            dns_blocking_events = []
            break

    for dns_be in dns_blocking_events:
        blocking_events.append(dns_be)

    if tcp_o_list:
        for tcp_o in tcp_o_list:
            observation_ids.append(tcp_o.observation_id)
            assert (
                domain_name == tcp_o.domain_name
            ), f"Inconsistent domain_name in tcp_o {tcp_o.domain_name}"
            tcp_ctrl = tcp_ctrl_map.get(f"{tcp_o.ip}:{tcp_o.port}")
            tcp_be = (
                make_website_tcp_blocking_event(tcp_o, tcp_ctrl) if tcp_ctrl else None
            )
            if tcp_be:
                blocking_events.append(tcp_be)

    if tls_o_list:
        for tls_o in tls_o_list:
            observation_ids.append(tls_o.observation_id)
            assert (
                domain_name == tls_o.domain_name
            ), f"Inconsistent domain_name in tls_o {tls_o.domain_name}"
            tls_be = make_website_tls_blocking_event(tls_o, blocking_events)
            if tls_be:
                blocking_events.append(tls_be)

    if http_o_list:
        for http_o in http_o_list:
            observation_ids.append(http_o.observation_id)
            assert (
                domain_name == http_o.domain_name
            ), f"Inconsistent domain_name in http_o {http_o.domain_name}"
            http_ctrl = http_ctrl_map.get(http_o.request_url)
            http_be = (
                make_website_http_blocking_event(
                    http_o, http_ctrl, blocking_events, fingerprintdb
                )
                if http_ctrl
                else None
            )
            if http_be:
                blocking_events.append(http_be)

    # TODO: Should we be including this also BlockingType.DOWN,SERVER_SIDE_BLOCK or not?
    ok_confidence = 1 - max(
        filter(
            lambda be: be.blocking_type
            not in (BlockingType.OK, BlockingType.DOWN, BlockingType.SERVER_SIDE_BLOCK),
            blocking_events,
        )
    )

    confirmed = False
    anomaly = False
    if ok_confidence == 0:
        confirmed = True
    if ok_confidence > 0.6:
        anomaly = True

    return WebsiteExperimentResult(
        blocking_events=blocking_events,
        observation_ids=observation_ids,
        anomaly=anomaly,
        confirmed=confirmed,
        ok_confidence=ok_confidence,
        **make_base_result_meta(nt_o),
    )



import io
import gzip
import itertools
import tarfile
import logging
import lz4.frame
import shutil
import orjson
import multiprocessing
import multiprocessing.pool
from pathlib import Path
from datetime import date, timedelta, datetime
from collections import defaultdict
from functools import partial

from dataclasses import dataclass

from enum import Enum
from typing import Callable, Generator, Set, List, Optional, NamedTuple, Union

import boto3
from botocore import UNSIGNED as botoSigUNSIGNED
from botocore.config import Config as botoConfig

from tqdm.contrib.logging import tqdm_logging_redirect

from oonidata.datautils import trim_measurement
from oonidata.normalize import iter_yaml_msmt_normalized
from oonidata.dataformat import trivial_id

LEGACY_BUCKET_NAME = "ooni-data"
MC_BUCKET_NAME = "ooni-data-eu-fra"

log = logging.getLogger("oonidata.dataclient")


def create_s3_client():
    return boto3.client("s3", config=botoConfig(signature_version=botoSigUNSIGNED))


s3 = create_s3_client()


def date_interval(start_day: date, end_day: date):
    """
    A generator for a date_interval.

    The end_day is not included in the range.
    """
    if start_day > end_day:
        return
    for d in range((end_day - start_day).days):
        yield start_day + timedelta(days=d)


def get_file_entry_ext(filename: str):
    supported_extensions = [
        # New jsonl gzips
        "jsonl.gz",
        # Old json gzips
        "json.gz",
        # New postcans
        "tar.gz",
        # Old cans
        "tar.lz4",
        "json.lz4",
        "yaml.lz4",
    ]
    for ext in supported_extensions:
        if filename.endswith("." + ext):
            return ext
    return filename.split(".")[-1]


class FilenameMeta(NamedTuple):
    timestamp: datetime
    probe_cc: Optional[str]
    testname: str


def get_filename_meta_v2(filename: str) -> FilenameMeta:
    p = filename.split("_")
    ts = p[0]
    # Some older timestamps don't have the hour in them, so we add the hour back
    # into it and set it to midnight.
    if len(ts) == 8:
        ts += "00"

    return FilenameMeta(
        timestamp=datetime.strptime(ts, "%Y%m%d%H"),
        probe_cc=p[1],
        testname=p[2].split(".")[0],
    )


def get_filename_meta_tarcan(s3path: str, filename: str) -> FilenameMeta:
    """
    These look like this:
    canned/2020-10-21/web_connectivity.0.tar.lz4
    """
    return FilenameMeta(
        timestamp=datetime.strptime(s3path.split("/")[-2], "%Y-%m-%d"),
        # These test_names have an underscore, we uniform them to the new format
        # without the underscore
        testname=filename.split(".")[0].replace("_", ""),
        # We don't know the probe_cc in these cases
        probe_cc=None,
    )


def get_filename_meta_reportcan(s3path: str, filename: str) -> FilenameMeta:
    """
    These look like this:
    canned/2020-10-21/20201021T141726Z-US-AS20093-facebook_messenger-20201021T141726Z_AS20093_MBCoLaOwlb5pq7CWYvsqhN5qc1Nd3rHQC0xmvoG1bCTn5ZCpS6-0.2.0-probe.json.lz4
    """
    p = filename.split("-")
    return FilenameMeta(
        # Note: we don't use the timestamp inside of the filename, because that
        # is the timestamp of the measurement, not of the measurement upload and
        # publication time.
        timestamp=datetime.strptime(s3path.split("/")[-2], "%Y-%m-%d"),
        # These test_names have an underscore, we uniform them to the new format
        # without the underscore
        testname=p[3].replace("_", ""),
        probe_cc=p[1],
    )


def read_to_bytesio(body: io.BytesIO) -> io.BytesIO:
    read_body = io.BytesIO()
    shutil.copyfileobj(body, read_body)
    read_body.seek(0)
    return read_body


def stream_jsonl(body: io.BytesIO) -> Generator[dict, None, None]:
    with gzip.GzipFile(fileobj=body) as in_file:
        for line in in_file:
            if line == "":
                continue
            yield orjson.loads(line)


def stream_postcan(body: io.BytesIO) -> Generator[dict, None, None]:
    # Since some older postcans have the .gz extension, but are actually not
    # compressed, tar needs to be able to re-seek back to the beginning of the
    # file in the event of it not finding the gzip magic header when operating
    # in "transparent compression mode".
    # When we we fix that in the source data, we might be able to avoid this.
    read_body = read_to_bytesio(body)

    with tarfile.open(fileobj=read_body) as tar:
        for m in tar:
            if not m.name.endswith(".post"):
                log.error(f"invalid filename in tar {m.name}")
                continue
            in_file = tar.extractfile(m)
            if in_file is None:
                log.error(f"empty file in tar {m.name}")
                continue

            post = orjson.loads(in_file.read())
            fmt = post.get("format", "")
            if fmt == "json":
                msmt = post.get("content", {})
                # extract msmt_uid from filename e.g:
                # ... /20210614004521.999962_JO_signal_68eb19b439326d60.post
                msmt_uid = m.name.rsplit("/", 1)[1]
                msmt_uid = msmt_uid[:-5]
                msmt["measurement_uid"] = msmt_uid
                yield msmt

            elif fmt == "yaml":
                log.info("Skipping YAML")

            else:
                log.error("Ignoring invalid post")


def stream_jsonlz4(body: io.BytesIO):
    # lz4.frame requires the input stream to be seekable, so we need to load it
    # in memory
    read_body = read_to_bytesio(body)

    with lz4.frame.open(read_body, mode="rb") as in_file:
        for line in in_file:
            try:
                msmt = orjson.loads(line)
            except ValueError:
                log.error("oldcan: unable to parse json measurement")
                continue

            msmt_uid = trivial_id(line, msmt)  # type: ignore due to bad types in lz4
            msmt["measurement_uid"] = msmt_uid
            yield msmt


def stream_yamllz4(body: io.BytesIO, s3path: str):
    # lz4.frame requires the input stream to be seekable, so we need to load it
    # in memory
    read_body = read_to_bytesio(body)

    with lz4.frame.open(read_body) as in_file:
        bucket_tstamp = s3path.split("/")[-2]
        rfn = f"{bucket_tstamp}/" + s3path.split("/")[-1]
        # The normalize function already add the measurement_uid
        yield from iter_yaml_msmt_normalized(in_file, bucket_tstamp, rfn)


def stream_oldcan(body: io.BytesIO, s3path: str) -> Generator[dict, None, None]:
    # lz4.frame requires the input stream to be seekable, so we need to load it
    # in memory
    read_body = read_to_bytesio(body)

    with lz4.frame.open(read_body) as lz4_file:
        with tarfile.open(fileobj=lz4_file) as tar:  # type: ignore due to bad types in lz4
            for m in tar:
                in_file = tar.extractfile(m)
                if in_file is None:
                    log.error(f"empty file in tar {m.name}")
                    continue

                if m.name.endswith(".json"):
                    for line in in_file:
                        try:
                            msmt = orjson.loads(line)
                        except ValueError:
                            log.error("oldcan: unable to parse json measurement")
                            continue

                        msmt_uid = trivial_id(line, msmt)
                        msmt["measurement_uid"] = msmt_uid
                        yield msmt

                elif m.name.endswith(".yaml"):
                    bucket_tstamp = s3path.split("/")[-2]
                    rfn = f"{bucket_tstamp}/" + s3path.split("/")[-1]
                    # The normalize function already add the measurement_uid
                    yield from iter_yaml_msmt_normalized(in_file, bucket_tstamp, rfn)


@dataclass
class FileEntry:
    s3path: str
    bucket_name: str
    timestamp: datetime
    testname: str
    filename: str
    size: int
    ext: str
    is_can: bool
    probe_cc: Optional[str] = None

    def matches_filter(
        self,
        ccs: Set[str],
        testnames: Set[str],
        start_timestamp: datetime,
        end_timestamp: datetime,
    ) -> bool:
        if self.probe_cc and ccs and self.probe_cc not in ccs:
            return False

        if self.testname and testnames and self.testname not in testnames:
            return False

        if self.timestamp < start_timestamp or self.timestamp >= end_timestamp:
            return False

        return True

    def stream_measurements(self):
        body = s3.get_object(Bucket=self.bucket_name, Key=self.s3path)["Body"]
        log.debug(f"streaming file {self}")
        if self.ext == "jsonl.gz":
            yield from stream_jsonl(body)
        elif self.ext == "tar.gz":
            yield from stream_postcan(body)
        elif self.ext == "tar.lz4":
            yield from stream_oldcan(body, self.s3path)
        elif self.ext == "json.lz4":
            yield from stream_jsonlz4(body)
        elif self.ext == "yaml.lz4":
            yield from stream_jsonlz4(body)
        else:
            log.error(
                f"found a file with an unknown extension: {self.ext} s3://{self.bucket_name}/{self.s3path}"
            )

    @staticmethod
    def from_obj_dict(bucket_name: str, obj_dict: dict) -> "FileEntry":
        s3path = obj_dict["Key"]
        filename = s3path.split("/")[-1]
        ext = get_file_entry_ext(filename)

        if ext == "jsonl.gz" or ext == "tar.gz":
            fm = get_filename_meta_v2(filename)
            is_can = False
            if ext == "tar.gz":
                is_can = True
        elif ext == "tar.lz4":
            fm = get_filename_meta_tarcan(s3path, filename)
            is_can = True
        elif ext == "json.lz4" or ext == "yaml.lz4":
            fm = get_filename_meta_reportcan(s3path, filename)
            is_can = True
        else:
            raise ValueError(f"unsupported file entry extension: {ext}")

        return FileEntry(
            s3path=s3path,
            bucket_name=bucket_name,
            timestamp=fm.timestamp,
            testname=fm.testname,
            filename=filename,
            size=obj_dict["Size"],
            ext=ext,
            probe_cc=fm.probe_cc,
            is_can=is_can,
        )


def list_all_testnames() -> Set[str]:
    testnames = set()
    paginator = s3.get_paginator("list_objects_v2")
    for r in paginator.paginate(Bucket=MC_BUCKET_NAME, Prefix="jsonl/", Delimiter="/"):
        for f in r.get("CommonPrefixes", []):
            testnames.add(f["Prefix"].split("/")[-2])
    return testnames


class Prefix(NamedTuple):
    bucket_name: str
    prefix: str


def get_v2_search_prefixes(testnames: Set[str], ccs: Set[str]) -> List[Prefix]:
    """
    get_search_prefixes will return all the prefixes inside of the new jsonl
    bucket that match the given testnames and ccs.
    If the ccs list is empty we will return prefixes for all countries for
    which that particular testname as measurements.
    """
    prefixes = []
    paginator = s3.get_paginator("list_objects_v2")
    for tn in testnames:
        for r in paginator.paginate(
            Bucket=MC_BUCKET_NAME, Prefix=f"jsonl/{tn}/", Delimiter="/"
        ):
            for f in r.get("CommonPrefixes", []):
                prefix = f["Prefix"]
                cc = prefix.split("/")[-2]
                if ccs and cc not in ccs:
                    continue
                prefixes.append(prefix)
    return prefixes


def get_v2_prefixes(
    ccs: Set[str], testnames: Set[str], start_day: date, end_day: date
) -> List[Prefix]:
    legacy_prefixes = [
        Prefix(bucket_name=MC_BUCKET_NAME, prefix=f"raw/{d:%Y%m%d}")
        for d in date_interval(max(date(2020, 10, 20), start_day), end_day)
    ]
    if not testnames:
        testnames = list_all_testnames()
    prefixes = []
    if start_day < date(2020, 10, 21):
        prefixes = get_v2_search_prefixes(testnames, ccs)
        combos = list(itertools.product(prefixes, date_interval(start_day, end_day)))
        # This results in a faster listing in cases where we need only a small time
        # window or few testnames. For larger windows of time, we are better off
        # just listing everything.
        if len(combos) < 1_000_000:  # XXX we might want to tweak this parameter a bit
            prefixes = [
                Prefix(bucket_name=MC_BUCKET_NAME, prefix=f"{p}{d:%Y%m%d}")
                for p, d in combos
            ]

    return prefixes + legacy_prefixes


def get_can_prefixes(start_day: date, end_day: date) -> List[Prefix]:
    """
    Returns the list of search prefixes for cans. In most cases, since we don't
    have the country code or test name in the path, all we do is return the
    range of dates.
    """
    new_cans = [
        Prefix(prefix=f"canned/{d:%Y-%m-%d}", bucket_name=MC_BUCKET_NAME)
        for d in date_interval(
            # The new cans are between 2020-06-02 and 2020-10-21 inclusive
            max(date(2020, 6, 2), start_day),
            min(date(2020, 10, 22), end_day),
        )
    ]
    old_cans = [
        Prefix(prefix=f"canned/{d:%Y-%m-%d}", bucket_name=LEGACY_BUCKET_NAME)
        for d in date_interval(
            # The new cans are between 2020-06-02 and 2020-10-21 inclusive
            # Note: the cans between 2020-06-02 and 2020-10-21 appears to be duplicated between the new and old cans.
            # TODO: check if they are actually identical or not.
            max(date(2012, 12, 5), start_day),
            min(date(2020, 6, 2), end_day),
        )
    ]
    return old_cans + new_cans


def iter_file_entries(prefix: Prefix) -> Generator[FileEntry, None, None]:
    s3_client = create_s3_client()
    paginator = s3_client.get_paginator("list_objects_v2")
    for r in paginator.paginate(Bucket=prefix.bucket_name, Prefix=prefix.prefix):
        for obj_dict in r.get("Contents", []):
            try:
                yield FileEntry.from_obj_dict(prefix.bucket_name, obj_dict)
            except ValueError as exc:
                log.error(exc)


def list_file_entries(prefix: Prefix) -> List[FileEntry]:
    return [fe for fe in iter_file_entries(prefix)]


class ProgressStatus(Enum):
    LISTING = "listing"
    DOWNLOAD_BEGIN = "download_begin"
    DOWNLOADING = "downloading"


class MeasurementListProgress(NamedTuple):
    current_prefix_idx: int
    total_prefixes: int

    current_file_entry_idx: int
    total_file_entries: int

    progress_status: ProgressStatus

    current_file_entry_bytes: int
    total_file_entry_bytes: int


def make_listing_progress(
    current_prefix_idx: int, total_prefixes: int, total_file_entries: int
):
    return MeasurementListProgress(
        current_prefix_idx=current_prefix_idx,
        total_prefixes=total_prefixes,
        progress_status=ProgressStatus.LISTING,
        current_file_entry_idx=0,
        total_file_entries=total_file_entries,
        current_file_entry_bytes=0,
        total_file_entry_bytes=0,
    )


def make_download_progress(
    current_file_entry_bytes: int,
    total_file_entry_bytes: int,
    current_file_entry_idx: int,
    total_file_entries: int,
    progress_status: ProgressStatus = ProgressStatus.DOWNLOADING,
):
    return MeasurementListProgress(
        current_file_entry_idx=current_file_entry_idx,
        total_file_entries=total_file_entries,
        current_file_entry_bytes=current_file_entry_bytes,
        total_file_entry_bytes=total_file_entry_bytes,
        progress_status=progress_status,
        total_prefixes=0,
        current_prefix_idx=0,
    )


CSVList = Optional[Union[List[str], str]]


def testnames_set(test_name: CSVList) -> Set[str]:
    if test_name is not None:
        if isinstance(test_name, str):
            test_name = test_name.split(",")
        return set(list(map(lambda x: x.lower().replace("_", ""), test_name)))
    return set()


def ccs_set(probe_cc: CSVList) -> Set[str]:
    if probe_cc is not None:
        if isinstance(probe_cc, str):
            probe_cc = probe_cc.split(",")
        return set(list(map(lambda x: x.upper(), probe_cc)))
    return set()


def get_file_entries(
    start_day: date,
    end_day: date,
    probe_cc: CSVList,
    test_name: CSVList,
    from_cans: bool,
    progress_callback: Optional[Callable[[MeasurementListProgress], None]] = None,
) -> List[FileEntry]:

    ccs = ccs_set(probe_cc)
    testnames = testnames_set(test_name)

    start_timestamp = datetime.combine(start_day, datetime.min.time())
    end_timestamp = datetime.combine(end_day, datetime.min.time())

    prefix_list = get_v2_prefixes(ccs, testnames, start_day, end_day)
    if from_cans == True:
        prefix_list = get_can_prefixes(start_day, end_day) + prefix_list

    log.debug(f"using prefix list {prefix_list}")
    file_entries = []
    prefix_idx = 0
    total_prefixes = len(prefix_list)

    if progress_callback:
        progress_callback(
            make_listing_progress(
                current_prefix_idx=0,
                total_prefixes=total_prefixes,
                total_file_entries=0,
            )
        )
    with multiprocessing.pool.ThreadPool() as pool:
        for fe_list in pool.imap_unordered(list_file_entries, prefix_list):
            for fe in fe_list:
                if not fe.matches_filter(
                    ccs, testnames, start_timestamp, end_timestamp
                ):
                    continue

                if from_cans == True and not fe.is_can:
                    continue
                if from_cans == False and fe.is_can:
                    continue

                file_entries.append(fe)
            prefix_idx += 1
            if progress_callback:
                progress_callback(
                    make_listing_progress(
                        current_prefix_idx=prefix_idx,
                        total_prefixes=total_prefixes,
                        total_file_entries=len(file_entries),
                    )
                )

    return file_entries


def iter_measurements(
    start_day: Union[date, str],
    end_day: Union[date, str],
    probe_cc: CSVList = None,
    test_name: CSVList = None,
    from_cans: bool = True,
    file_entries: Optional[List[FileEntry]] = None,
    progress_callback: Optional[Callable[[MeasurementListProgress], None]] = None,
) -> Generator[dict, None, None]:
    ccs = ccs_set(probe_cc)

    if isinstance(start_day, str):
        start_day = datetime.strptime(start_day, "%Y-%m-%d").date()
    if isinstance(end_day, str):
        end_day = datetime.strptime(end_day, "%Y-%m-%d").date()

    if file_entries is None:
        file_entries = get_file_entries(
            start_day=start_day,
            end_day=end_day,
            test_name=test_name,
            probe_cc=probe_cc,
            from_cans=from_cans,
            progress_callback=progress_callback,
        )

    total_file_entry_bytes = sum(map(lambda fe: fe.size, file_entries))
    if progress_callback:
        progress_callback(
            make_download_progress(
                current_file_entry_bytes=0,
                total_file_entry_bytes=total_file_entry_bytes,
                current_file_entry_idx=0,
                total_file_entries=len(file_entries),
                progress_status=ProgressStatus.DOWNLOAD_BEGIN,
            )
        )

    for idx, fe in enumerate(file_entries):
        for msmt in fe.stream_measurements():
            # Legacy cans don't allow us to pre-filter on the probe_cc, so we do
            # it before returning the data to the caller
            if ccs and msmt["probe_cc"] not in ccs:
                continue
            yield msmt

        if progress_callback:
            progress_callback(
                make_download_progress(
                    current_file_entry_bytes=fe.size,
                    total_file_entry_bytes=total_file_entry_bytes,
                    current_file_entry_idx=idx,
                    total_file_entries=len(file_entries),
                )
            )


def make_filename(max_string_size: Optional[int], fe: FileEntry) -> str:
    flags = ""
    if max_string_size:
        flags = f"_max{max_string_size}"
    ts = fe.timestamp.strftime("%Y%m%d%H")
    filename = f"{ts}_{fe.probe_cc}_{fe.testname}{flags}.jsonl.gz"
    return filename


def download_file_entry_list(
    fe_list: List[FileEntry],
    output_dir: Path,
    probe_cc: List[str],
    test_name: List[str],
    start_day: date,
    end_day: date,
    max_string_size: Optional[int],
):
    """
    Download a list of file entries to the output dir.

    It assumes that the list of file entries in the list are all pertinent to
    the same testname, probe_cc, hour tuple
    """
    total_fe_size = 0
    output_dir = (
        output_dir / fe_list[0].testname / fe_list[0].timestamp.strftime("%Y-%m-%d")
    )
    output_dir.mkdir(parents=True, exist_ok=True)

    output_path = output_dir / make_filename(max_string_size, fe_list[0])

    with gzip.open(output_path.with_suffix(".tmp"), "wb") as out_file:
        for fe in fe_list:
            assert fe.testname == fe_list[0].testname
            assert fe.timestamp == fe_list[0].timestamp
            assert fe.probe_cc == fe_list[0].probe_cc
            total_fe_size += fe.size

        for msmt_dict in iter_measurements(
            start_day=start_day,
            end_day=end_day,
            probe_cc=probe_cc,
            test_name=test_name,
            file_entries=fe_list,
        ):
            if max_string_size:
                msmt_dict = trim_measurement(msmt_dict, max_string_size)
            out_file.write(orjson.dumps(msmt_dict))
            out_file.write(b"\n")

    output_path.with_suffix(".tmp").rename(output_path)
    return total_fe_size


def sync_measurements(
    output_dir: Path,
    probe_cc: List[str],
    test_name: List[str],
    start_day: date,
    end_day: date,
    max_string_size: Optional[int] = None,
):
    with tqdm_logging_redirect(unit_scale=True) as pbar:

        def cb_list_fe(p):
            if p.current_prefix_idx == 0:
                pbar.total = p.total_prefixes
                pbar.update(0)
                pbar.set_description("starting prefix listing")
                return
            pbar.update(1)
            pbar.set_description(
                f"listed {p.total_file_entries} files in {p.current_prefix_idx}/{p.total_prefixes} prefixes"
            )

        all_file_entries = get_file_entries(
            start_day=start_day,
            end_day=end_day,
            test_name=test_name,
            probe_cc=probe_cc,
            from_cans=True,
            progress_callback=cb_list_fe,
        )

        total_fe_size = 0
        to_download_fe = defaultdict(list)
        for fe in all_file_entries:
            dst_path = (
                output_dir
                / fe.testname
                / fe.timestamp.strftime("%Y-%m-%d")
                / make_filename(max_string_size, fe)
            )
            if dst_path.exists():
                continue

            ts = fe.timestamp.strftime("%Y%m%d%H")
            # We group based on this key, so each process is writing to the same file.
            # Each raw folder can have multiple files on a given hour
            key = f"{ts}-{fe.testname}-{fe.probe_cc}"
            to_download_fe[key].append(fe)
            total_fe_size += fe.size

        pbar.unit = "B"
        pbar.reset(total=total_fe_size)
        pbar.set_description("downloading files")
        download_count = 0
        with multiprocessing.Pool() as pool:
            for fe_size in pool.imap_unordered(
                partial(
                    download_file_entry_list,
                    output_dir=output_dir,
                    probe_cc=probe_cc,
                    test_name=test_name,
                    start_day=start_day,
                    end_day=end_day,
                    max_string_size=max_string_size,
                ),
                to_download_fe.values(),
            ):
                download_count += 1
                pbar.update(fe_size)
                pbar.set_description(
                    f"downloaded {download_count}/{len(to_download_fe)}"
                )



from datetime import datetime, date, timedelta
import logging
from typing import Iterable, List, Any, Tuple, Union, Dict
from dataclasses import dataclass
from functools import partial, singledispatch
import re
import ipaddress

from base64 import b64decode

from OpenSSL.crypto import load_certificate, FILETYPE_PEM, FILETYPE_ASN1
from OpenSSL.crypto import X509Store, X509StoreContext

from cryptography import x509
from cryptography.x509.oid import ExtensionOID, NameOID
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization


from oonidata.dataformat import (
    HeadersListBytes,
    HeadersListStr,
    BinaryData,
    guess_decode,
)


log = logging.getLogger("oonidata.datautils")

META_TITLE_REGEXP = re.compile(
    b'<meta.*?property="og:title".*?content="(.*?)"', re.IGNORECASE | re.DOTALL
)


def get_html_meta_title(body: bytes) -> str:
    m = META_TITLE_REGEXP.search(body, re.IGNORECASE | re.DOTALL)
    if m:
        return guess_decode(m.group(1))
    return ""


TITLE_REGEXP = re.compile(b"<title.*?>(.*?)</title>", re.IGNORECASE | re.DOTALL)


def get_html_title(body: bytes) -> str:
    m = META_TITLE_REGEXP.search(body, re.IGNORECASE | re.DOTALL)
    if m:
        return guess_decode(m.group(1))
    return ""


def get_first_http_header(
    header_name: str,
    header_list: HeadersListBytes,
    case_sensitive: bool = False,
) -> bytes:
    if not header_list:
        return b""

    if case_sensitive == False:
        header_name = header_name.lower()

    for k, v in header_list:
        if case_sensitive == False:
            k = k.lower()

        if header_name == k:
            return v

    return b""


# TODO: come up with some nicer way to refactor this stuff so we don't have so much duplicate functions
def get_first_http_header_str(
    header_name: str,
    header_list: HeadersListStr,
    case_sensitive: bool = False,
) -> Union[bytes, str, None]:
    if not header_list:
        return ""

    if case_sensitive == False:
        header_name = header_name.lower()

    for k, v in header_list:
        if case_sensitive == False:
            k = k.lower()

        if header_name == k:
            return v

    return ""


# This comes from: https://ipinfo.io/bogon and https://publicdata.caida.org/datasets/bogon/bogon-bn-agg/
bogon_ipv4_ranges = [
    ipaddress.ip_network("0.0.0.0/8"),  # "This" network
    ipaddress.ip_network("10.0.0.0/8"),  # Private-use networks
    ipaddress.ip_network("100.64.0.0/10"),  # Carrier-grade NAT
    ipaddress.ip_network("127.0.0.0/8"),  # Loopback
    ipaddress.ip_network("169.254.0.0/16"),  # Link local
    ipaddress.ip_network("172.16.0.0/12"),  # Private-use networks
    ipaddress.ip_network("192.0.0.0/24"),  # IETF protocol assignments
    ipaddress.ip_network("192.0.2.0/24"),  # TEST-NET-1
    ipaddress.ip_network("192.168.0.0/16"),  # Private-use networks
    # Network interconnect device benchmark testing
    ipaddress.ip_network("198.18.0.0/15"),
    ipaddress.ip_network("198.51.100.0/24"),  # TEST-NET-2
    ipaddress.ip_network("203.0.113.0/24"),  # TEST-NET-3
    ipaddress.ip_network("224.0.0.0/3"),  # Multicast
]


bogon_ipv6_ranges = [
    # Node-scope unicast unspecified address
    ipaddress.ip_network("::/128"),
    # Node-scope unicast loopback address
    ipaddress.ip_network("::1/128"),
    # IPv4-mapped addresses
    ipaddress.ip_network("::ffff:0:0/96"),
    # IPv4-compatible addresses
    ipaddress.ip_network("::/96"),
    # Remotely triggered black hole addresses
    ipaddress.ip_network("100::/64"),
    # Overlay routable cryptographic hash identifiers (ORCHID)
    ipaddress.ip_network("2001:10::/28"),
    # Documentation prefix
    ipaddress.ip_network("2001:db8::/32"),
    # Unique local addresses (ULA)
    ipaddress.ip_network("fc00::/7"),
    # Link-local unicast
    ipaddress.ip_network("fe80::/10"),
    # Site-local unicast (deprecated)
    ipaddress.ip_network("fec0::/10"),
    # Multicast (Note: ff0e:/16 is global scope and may appear on the global internet.)
    ipaddress.ip_network("ff00::/8"),
    # 6to4 bogon (0.0.0.0/8)
    ipaddress.ip_network("2002::/24"),
    # 6to4 bogon (10.0.0.0/8)
    ipaddress.ip_network("2002:a00::/24"),
    # 6to4 bogon (127.0.0.0/8)
    ipaddress.ip_network("2002:7f00::/24"),
    # 6to4 bogon (169.254.0.0/16)
    ipaddress.ip_network("2002:a9fe::/32"),
    # 6to4 bogon (172.16.0.0/12)
    ipaddress.ip_network("2002:ac10::/28"),
    # 6to4 bogon (192.0.0.0/24)
    ipaddress.ip_network("2002:c000::/40"),
    # 6to4 bogon (192.0.2.0/24)
    ipaddress.ip_network("2002:c000:200::/40"),
    # 6to4 bogon (192.168.0.0/16)
    ipaddress.ip_network("2002:c0a8::/32"),
    # 6to4 bogon (198.18.0.0/15)
    ipaddress.ip_network("2002:c612::/31"),
    # 6to4 bogon (198.51.100.0/24)
    ipaddress.ip_network("2002:c633:6400::/40"),
    # 6to4 bogon (203.0.113.0/24)
    ipaddress.ip_network("2002:cb00:7100::/40"),
    # 6to4 bogon (224.0.0.0/4)
    ipaddress.ip_network("2002:e000::/20"),
    # 6to4 bogon (240.0.0.0/4)
    ipaddress.ip_network("2002:f000::/20"),
    # 6to4 bogon (255.255.255.255/32)
    ipaddress.ip_network("2002:ffff:ffff::/48"),
    # Teredo bogon (0.0.0.0/8)
    ipaddress.ip_network("2001::/40"),
    # Teredo bogon (10.0.0.0/8)
    ipaddress.ip_network("2001:0:a00::/40"),
    # Teredo bogon (127.0.0.0/8)
    ipaddress.ip_network("2001:0:7f00::/40"),
    # Teredo bogon (169.254.0.0/16)
    ipaddress.ip_network("2001:0:a9fe::/48"),
    # Teredo bogon (172.16.0.0/12)
    ipaddress.ip_network("2001:0:ac10::/44"),
    # Teredo bogon (192.0.0.0/24)
    ipaddress.ip_network("2001:0:c000::/56"),
    # Teredo bogon (192.0.2.0/24)
    ipaddress.ip_network("2001:0:c000:200::/56"),
    # Teredo bogon (192.168.0.0/16)
    ipaddress.ip_network("2001:0:c0a8::/48"),
    # Teredo bogon (198.18.0.0/15)
    ipaddress.ip_network("2001:0:c612::/47"),
    # Teredo bogon (198.51.100.0/24)
    ipaddress.ip_network("2001:0:c633:6400::/56"),
    # Teredo bogon (203.0.113.0/24)
    ipaddress.ip_network("2001:0:cb00:7100::/56"),
    # Teredo bogon (224.0.0.0/4)
    ipaddress.ip_network("2001:0:e000::/36"),
    # Teredo bogon (240.0.0.0/4)
    ipaddress.ip_network("2001:0:f000::/36"),
    # Teredo bogon (255.255.255.255/32)
    ipaddress.ip_network("2001:0:ffff:ffff::/64"),
]


def is_ipv4_bogon(ip: str) -> bool:
    try:
        ipv4addr = ipaddress.IPv4Address(ip)
    except ipaddress.AddressValueError:
        return False
    if any([ipv4addr in ip_range for ip_range in bogon_ipv4_ranges]):
        return True
    return False


def is_ipv6_bogon(ip: str) -> bool:
    try:
        ipv6addr = ipaddress.IPv6Address(ip)
    except ipaddress.AddressValueError:
        return False
    if any([ipv6addr in ip_range for ip_range in bogon_ipv6_ranges]):
        return True
    return False


def is_ip_bogon(ip: str) -> bool:
    ipaddr = ipaddress.ip_address(ip)
    if any([ipaddr in ip_range for ip_range in bogon_ipv4_ranges + bogon_ipv6_ranges]):
        return True
    return False


@dataclass
class CertificateMeta:
    cert: x509.Certificate
    issuer: str
    issuer_common_name: str
    subject: str
    subject_common_name: str
    san_list: List[str]
    not_valid_before: datetime
    not_valid_after: datetime
    fingerprint: str


def get_common_name(cert_name: x509.Name) -> str:
    try:
        attributes = cert_name.get_attributes_for_oid(NameOID.COMMON_NAME)
        if attributes:
            return str(attributes[0].value)
    except x509.AttributeNotFound:
        return ""
    return ""


def get_san_list(cert: x509.Certificate) -> List[str]:
    try:
        ext = cert.extensions.get_extension_for_oid(
            ExtensionOID.SUBJECT_ALTERNATIVE_NAME
        )
        san_ext: x509.SubjectAlternativeName = ext.value  # type: ignore
        return san_ext.get_values_for_type(x509.DNSName)
    except x509.ExtensionNotFound:
        return []


def get_certificate_meta(peer_cert: BinaryData) -> CertificateMeta:
    raw_cert = b64decode(peer_cert.data)
    assert raw_cert, "either peer_cert or raw_cert must be specified"
    cert = x509.load_der_x509_certificate(raw_cert, default_backend())

    return CertificateMeta(
        cert=cert,
        issuer=cert.issuer.rfc4514_string(),
        issuer_common_name=get_common_name(cert.issuer),
        subject=cert.subject.rfc4514_string(),
        subject_common_name=get_common_name(cert.subject),
        san_list=get_san_list(cert),
        not_valid_before=cert.not_valid_before,
        not_valid_after=cert.not_valid_after,
        fingerprint=cert.fingerprint(hashes.SHA256()).hex(),
    )


class InvalidCertificateChain(ValueError):
    pass


class TLSCertStore:
    def __init__(self, pem_cert_store: Iterable[bytes] = []):
        self._store: Dict[x509.Name, x509.Certificate] = {}
        for pem_cert in pem_cert_store:
            self.add_cert_to_store(pem_cert)

    def add_cert_to_store(self, pem_cert: bytes):
        root_cert = x509.load_pem_x509_certificate(pem_cert)

        if root_cert.issuer in self._store:
            log.error(f"duplicate cert {root_cert.issuer} in store {self._store}")
            return

        self._store[root_cert.issuer] = root_cert

    def validate_cert_chain(
        self, timestamp: datetime, certificate_chain: List[bytes]
    ) -> Tuple[str, List[str]]:
        """
        Validate a certificate chain provided as a list of bytes in DER format
        against a set of trust anchors (`pem_cert_store`).
        The pem_cert_store should be a list of bytes of root certificates in PEM
        format.

        The user of this function needs to implement their own logic for verifying
        if the returned SAN list and common name matches the expectations.

        It's important to highlight that this function doesn't make any sort of
        strong security guarantees and doesn't take into account the many
        nuances and edge cases that MUST be considered if you care to have a
        secure certificate chain validation function as per:
        https://www.rfc-editor.org/rfc/rfc5280#section-6

        Returns:
            (common_name, List[SAN list])

        Raises:
            InvalidCertificateChain if the certificate is not valid
        """
        # The cert chain as `x509.Certificates` object. The first item in the list
        # is the leaf node of the chain.
        cert_chain = list(
            map(
                partial(x509.load_der_x509_certificate, backend=default_backend),
                certificate_chain,
            )
        )

        try:
            issuing_root = self._store[cert_chain[-1].issuer]
        except KeyError:
            raise InvalidCertificateChain(
                f"missing issuing_root {cert_chain[-1].issuer.rfc4514_string()} in cert store"
            )

        store = X509Store()
        store.set_time(timestamp)
        store.add_cert(
            load_certificate(
                FILETYPE_PEM,
                issuing_root.public_bytes(encoding=serialization.Encoding.PEM),
            )
        )
        for cert in reversed(certificate_chain):
            pcert = x509.load_der_x509_certificate(cert)
            if pcert.subject == issuing_root.issuer:
                # Skip certificates that are already in the store and hence are
                # considered trusted.
                # XXX: From a security perspective this might go wrong in some ways,
                # but we aren't really aiming for a secure implementation.
                continue

            unstrusted_cert = load_certificate(FILETYPE_ASN1, cert)
            store_ctx = X509StoreContext(store, unstrusted_cert)

            try:
                store_ctx.verify_certificate()
            except Exception as exc:
                log.error(f"failed to verify i={pcert.issuer} s={pcert.subject} {exc}")
                log.error(f"using i={issuing_root.issuer} s={issuing_root.subject}")
                raise InvalidCertificateChain(
                    f"failed to verify i={pcert.issuer} s={pcert.subject} {exc}"
                    f"using i={issuing_root.issuer} s={issuing_root.subject}"
                )

            store.add_cert(unstrusted_cert)

        san_list = get_san_list(cert_chain[0])
        common_name = get_common_name(cert_chain[0].subject)
        return common_name, san_list


def validate_cert_chain(
    timestamp: datetime,
    certificate_chain: List[bytes],
    pem_cert_store: Iterable[bytes],
) -> Tuple[str, List[str]]:
    store = TLSCertStore(pem_cert_store)
    return store.validate_cert_chain(timestamp, certificate_chain)


# Taken from:
# https://github.com/Jigsaw-Code/net-analysis/blob/master/netanalysis/ooni/data/sync_measurements.py#L33
@singledispatch
def trim_measurement(json_obj, max_string_size: int):
    return json_obj


@trim_measurement.register(dict)
def _(json_dict: dict, max_string_size: int):
    keys_to_delete: List[str] = []
    for key, value in json_dict.items():
        if type(value) == str and len(value) > max_string_size:
            keys_to_delete.append(key)
        else:
            trim_measurement(value, max_string_size)
    for key in keys_to_delete:
        del json_dict[key]
    return json_dict


@trim_measurement.register(list)
def _(json_list: list, max_string_size: int):
    for item in json_list:
        trim_measurement(item, max_string_size)
    return json_list


def one_day_dict(day: date) -> Dict[str, Any]:
    start_day = datetime(year=day.year, month=day.month, day=day.day)
    end_day = start_day + timedelta(days=1)
    return {"start_day": start_day, "end_day": end_day}



from base64 import b64decode
import hashlib
import abc
import logging

from dataclasses import dataclass, field
from urllib.parse import urlparse, urlsplit
from datetime import datetime, timedelta
from typing import Callable, Generator, Optional, List, Tuple, Union, Dict
from oonidata.dataformat import SIGNAL_PEM_STORE, DNSCheck, Tor

from oonidata.dataformat import (
    BaseMeasurement,
    DNSAnswer,
    DNSQuery,
    HTTPTransaction,
    Failure,
    NetworkEvent,
    Signal,
    TCPConnect,
    TLSHandshake,
    WebConnectivity,
)

from oonidata.datautils import (
    InvalidCertificateChain,
    TLSCertStore,
    get_first_http_header,
    get_first_http_header_str,
    get_html_meta_title,
    get_html_title,
    is_ipv4_bogon,
    is_ipv6_bogon,
    get_certificate_meta,
)
from oonidata.fingerprintdb import FingerprintDB
from oonidata.netinfo import NetinfoDB


log = logging.getLogger("oonidata.processing")


unknown_failure_map = (
    (
        "This is usually a temporary error during hostname resolution and means that the local server did not receive a response from an authoritative server",
        "dns_temporary_failure",
    ),
    (
        "Der angeforderte Name ist gültig, es wurden jedoch keine Daten des angeforderten Typs gefunden",
        "dns_temporary_failure",
    ),
    ("certificate has expired or is not yet valid", "ssl_invalid_certificate"),
)


def normalize_failure(failure: Failure):
    # TODO: implement a mapping between known unknowns to cleanup the data a bit
    if not failure:
        return failure

    for substring, new_failure in unknown_failure_map:
        if substring in failure:
            return new_failure
    return failure


@dataclass
class Observation(abc.ABC):
    __table_name__ = "obs"

    measurement_uid: str
    observation_id: str

    input: str
    report_id: str

    timestamp: datetime

    target: str

    probe_asn: int
    probe_cc: str

    probe_as_org_name: str
    probe_as_cc: str
    probe_as_name: str

    software_name: str
    software_version: str
    test_name: str
    test_version: str

    network_type: str
    platform: str
    origin: str

    resolver_ip: str
    resolver_asn: int
    resolver_cc: str
    resolver_as_org_name: str
    resolver_as_cc: str

    resolver_is_scrubbed: bool

    # This is the resolver metadata computed by the probe. Once we do some
    # quality control on them we might consolidate these into a single set of
    # fields.
    # If the resolver_is_scrubbed, we will be setting the resolver_* values to
    # those computed by the probe and resolver_ip will be the empty string.
    resolver_asn_probe: int
    resolver_as_org_name_probe: str


def make_base_observation_meta(msmt: BaseMeasurement, netinfodb: NetinfoDB) -> dict:
    assert msmt.measurement_uid is not None
    probe_asn = int(msmt.probe_asn[len("AS") :])
    measurement_start_time = datetime.strptime(
        msmt.measurement_start_time, "%Y-%m-%d %H:%M:%S"
    )
    probe_as_info = netinfodb.lookup_asn(measurement_start_time, probe_asn)

    resolver_as_info = None
    resolver_ip = msmt.resolver_ip
    client_resolver = None
    resolver_is_scrubbed = False
    if msmt.test_keys and msmt.test_keys.client_resolver:
        client_resolver = msmt.test_keys.client_resolver

    if client_resolver == "[scrubbed]" or resolver_ip == "[scrubbed]":
        resolver_is_scrubbed = True

    resolver_ip = resolver_ip or client_resolver or ""
    resolver_cc = ""
    resolver_asn = 0
    resolver_as_org_name = ""
    resolver_as_cc = ""

    resolver_asn_probe = msmt.resolver_asn
    if resolver_asn_probe is None:
        resolver_asn_probe = 0
    else:
        resolver_asn_probe = int(resolver_asn_probe[2:])
    resolver_as_org_name_probe = msmt.resolver_network_name or ""
    if resolver_ip == "[scrubbed]":
        resolver_asn = resolver_asn_probe
        resolver_as_org_name = resolver_as_org_name_probe
        resolver_ip = ""

    if resolver_ip != "":
        resolver_as_info = netinfodb.lookup_ip(measurement_start_time, resolver_ip)
        if resolver_as_info:
            resolver_cc = resolver_as_info.cc
            resolver_asn = resolver_as_info.as_info.asn
            resolver_as_org_name = resolver_as_info.as_info.as_org_name
            resolver_as_cc = resolver_as_info.as_info.as_cc

    input_ = msmt.input
    if isinstance(input_, list):
        input_ = ":".join(input_)

    return dict(
        measurement_uid=msmt.measurement_uid,
        probe_asn=probe_asn,
        probe_cc=msmt.probe_cc,
        probe_as_org_name=probe_as_info.as_org_name if probe_as_info else "",
        probe_as_cc=probe_as_info.as_cc if probe_as_info else "",
        probe_as_name=probe_as_info.as_name if probe_as_info else "",
        report_id=msmt.report_id,
        input=input_,
        software_name=msmt.software_name,
        software_version=msmt.software_version,
        test_name=msmt.test_name,
        test_version=msmt.test_version,
        network_type=msmt.annotations.get("network_type", "unknown"),
        platform=msmt.annotations.get("platform", "unknown"),
        origin=msmt.annotations.get("origin", "unknown"),
        resolver_ip=resolver_ip,
        resolver_cc=resolver_cc,
        resolver_asn=resolver_asn,
        resolver_as_org_name=resolver_as_org_name,
        resolver_as_cc=resolver_as_cc,
        resolver_asn_probe=resolver_asn_probe,
        resolver_as_org_name_probe=resolver_as_org_name_probe,
        resolver_is_scrubbed=resolver_is_scrubbed,
    )


def make_timestamp(msmt: BaseMeasurement, t: Optional[float] = None):
    timestamp = datetime.strptime(msmt.measurement_start_time, "%Y-%m-%d %H:%M:%S")
    if t:
        timestamp += timedelta(seconds=t)
    return timestamp


@dataclass
class NettestObservation(Observation):
    __table_name__ = "obs_nettest"

    test_runtime: float
    annotations: Dict[str, str]

    @staticmethod
    def from_measurement(
        msmt: BaseMeasurement,
        netinfodb: NetinfoDB,
    ) -> "NettestObservation":
        return NettestObservation(
            observation_id=f"{msmt.measurement_uid}_nettest",
            timestamp=make_timestamp(msmt),
            test_runtime=msmt.test_runtime,
            annotations=msmt.annotations,
            target="",
            **make_base_observation_meta(msmt, netinfodb),
        )


@dataclass
class HTTPObservation(Observation):
    __table_name__ = "obs_http"

    domain_name: str
    request_url: str

    network: str
    alpn: Optional[str]

    failure: Failure

    request_body_length: int
    # request_headers_list: Optional[List[Tuple[str, bytes]]]
    request_method: str

    response_fingerprints: List[str]

    ip: Optional[str] = None
    port: Optional[int] = None

    runtime: Optional[float] = None

    response_body_length: Optional[int] = None
    response_body_is_truncated: Optional[bool] = None
    response_body_sha1: Optional[str] = None
    response_body_title: Optional[str] = None
    response_body_meta_title: Optional[str] = None

    response_status_code: Optional[int] = None
    # response_headers_list: Optional[List[Tuple[str, bytes]]]
    response_header_location: Optional[bytes] = None
    response_header_server: Optional[bytes] = None
    request_redirect_from: Optional[str] = None
    request_body_is_truncated: Optional[bool] = None

    fingerprint_country_consistent: Optional[bool] = None
    response_matches_blockpage: bool = False
    response_matches_false_positive: bool = False

    @property
    def request_is_encrypted(self):
        return self.request_url.startswith("https://")

    @staticmethod
    def from_measurement(
        msmt: BaseMeasurement,
        netinfodb: NetinfoDB,
        idx: int,
        requests_list: Optional[List[HTTPTransaction]],
        http_transaction: HTTPTransaction,
        fingerprintdb: FingerprintDB,
        target: str = "",
    ) -> Optional["HTTPObservation"]:
        if not http_transaction.request:
            # This is a very malformed request, we don't consider it a valid
            # observation as we don't know what it's referring to.
            # XXX maybe log this somewhere
            return None

        network = (
            http_transaction.network or http_transaction.request.x_transport or "tcp"
        )
        # We uniform all observations to the new data format
        if network == "quic":
            network = "udp"

        parsed_url = urlparse(http_transaction.request.url)
        hrro = HTTPObservation(
            observation_id=f"{msmt.measurement_uid}_http_{idx}",
            request_url=http_transaction.request.url,
            domain_name=parsed_url.hostname or "",
            request_body_is_truncated=http_transaction.request.body_is_truncated,
            # hrro.request_headers_list = http_transaction.request.headers_list_bytes
            request_method=http_transaction.request.method or "",
            request_body_length=len(http_transaction.request.body_bytes)
            if http_transaction.request.body_bytes
            else 0,
            response_fingerprints=[],
            network=network,
            alpn=http_transaction.alpn,
            failure=normalize_failure(http_transaction.failure),
            timestamp=make_timestamp(msmt, http_transaction.t),
            target=target,
            **make_base_observation_meta(msmt, netinfodb),
        )

        if http_transaction.address:
            p = urlsplit("//" + http_transaction.address)
            hrro.ip = p.hostname
            hrro.port = p.port

        if http_transaction.t is not None and http_transaction.t0 is not None:
            hrro.runtime = http_transaction.t - http_transaction.t0

        if not http_transaction.response:
            return hrro

        hrro.response_body_is_truncated = http_transaction.response.body_is_truncated

        fp_matches = fingerprintdb.match_http(http_transaction.response)
        for fp in fp_matches:
            if fp.scope == "fp":
                hrro.response_matches_false_positive = True
            else:
                hrro.response_matches_blockpage = True
            if fp.expected_countries and msmt.probe_cc in fp.expected_countries:
                hrro.fingerprint_country_consistent = True
            hrro.response_fingerprints.append(fp.name)

        if http_transaction.response.body_bytes:
            hrro.response_body_length = len(http_transaction.response.body_bytes)
            hrro.response_body_sha1 = hashlib.sha1(
                http_transaction.response.body_bytes
            ).hexdigest()
            hrro.response_body_title = get_html_title(
                http_transaction.response.body_bytes
            )
            hrro.response_body_meta_title = get_html_meta_title(
                http_transaction.response.body_bytes
            )

        hrro.response_status_code = http_transaction.response.code
        # hrro.response_headers_list = http_transaction.response.headers_list_bytes

        hrro.response_header_location = get_first_http_header(
            "location", http_transaction.response.headers_list_bytes or []
        )
        hrro.response_header_server = get_first_http_header(
            "server", http_transaction.response.headers_list_bytes or []
        )

        try:
            # We add type: ignore in here, because requests_lists is an optional
            # field and the fact it might not be defined is handled by the
            # except block below, yet pylint is not able to figure that out.
            # TODO: maybe refactor this handle it better by checking if these are defined
            prev_request = requests_list[idx + 1]  # type: ignore
            prev_location = get_first_http_header_str(
                "location", prev_request.response.headers_list_str or []  # type: ignore
            )
            if prev_location == hrro.request_url:
                hrro.request_redirect_from = prev_request.request.url  # type: ignore
        except (IndexError, UnicodeDecodeError, AttributeError):
            pass
        return hrro


def make_http_observations(
    msmt: BaseMeasurement,
    requests_list: Optional[List[HTTPTransaction]],
    fingerprintdb: FingerprintDB,
    netinfodb: NetinfoDB,
    target: str = "",
) -> List[HTTPObservation]:
    obs_list = []
    if not requests_list:
        return obs_list

    for idx, http_transaction in enumerate(requests_list):
        httpo = HTTPObservation.from_measurement(
            msmt, netinfodb, idx, requests_list, http_transaction, fingerprintdb, target
        )
        if httpo:
            obs_list.append(httpo)
    return obs_list


@dataclass
class DNSObservation(Observation):
    __table_name__ = "obs_dns"

    domain_name: str

    query_type: str
    failure: Failure
    engine: Optional[str]
    engine_resolver_address: Optional[str]

    answer_type: Optional[str] = None
    answer: Optional[str] = None
    answer_asn: Optional[int] = None
    answer_as_org_name: Optional[str] = None
    answer_as_cc: Optional[str] = None
    answer_cc: Optional[str] = None
    answer_is_bogon: Optional[bool] = None

    fingerprint_id: Optional[str] = None
    fingerprint_country_consistent: Optional[bool] = None

    is_tls_consistent: Optional[bool] = None

    transaction_id: Optional[int] = None

    @staticmethod
    def from_measurement(
        msmt: BaseMeasurement,
        query: DNSQuery,
        answer: Optional[DNSAnswer],
        idx: int,
        fingerprintdb: FingerprintDB,
        netinfodb: NetinfoDB,
        target: str = "",
    ) -> "DNSObservation":
        dnso = DNSObservation(
            observation_id=f"{msmt.measurement_uid}_dns_{idx}",
            engine=query.engine,
            engine_resolver_address=query.resolver_address,
            query_type=query.query_type,
            domain_name=query.hostname,
            failure=normalize_failure(query.failure),
            timestamp=make_timestamp(msmt, query.t),
            target=target,
            transaction_id=query.transaction_id,
            **make_base_observation_meta(msmt, netinfodb),
        )

        if not answer:
            return dnso

        dnso.answer_type = answer.answer_type
        if answer.ipv4:
            dnso.answer = answer.ipv4
            dnso.answer_is_bogon = is_ipv4_bogon(answer.ipv4)
        elif answer.ipv6:
            dnso.answer = answer.ipv6
            dnso.answer_is_bogon = is_ipv6_bogon(answer.ipv6)
        elif answer.hostname:
            dnso.answer = answer.hostname

        if answer.ipv4 or answer.ipv6:
            # This is guaranteed to be the correct type since we set it's value
            # based on answer.ipv4 or answer.ipv6 being set in the previous
            # blocks
            answer_meta = netinfodb.lookup_ip(dnso.timestamp, dnso.answer)  # type: ignore
            if answer_meta:
                dnso.answer_asn = answer_meta.as_info.asn
                dnso.answer_as_cc = answer_meta.as_info.as_cc
                dnso.answer_as_org_name = answer_meta.as_info.as_org_name
                dnso.answer_cc = answer_meta.cc

        matched_fingerprint = fingerprintdb.match_dns(dnso.answer)
        if matched_fingerprint:
            dnso.fingerprint_id = matched_fingerprint.name
            if matched_fingerprint.expected_countries:
                dnso.fingerprint_country_consistent = (
                    msmt.probe_cc in matched_fingerprint.expected_countries
                )
        return dnso


def make_dns_observations(
    msmt: BaseMeasurement,
    queries: Optional[List[DNSQuery]],
    fingerprintdb: FingerprintDB,
    netinfodb: NetinfoDB,
    target: str = "",
) -> List[DNSObservation]:
    obs_dns = []
    if not queries:
        return obs_dns

    idx = 0
    for query in queries:
        answer_list = query.answers
        if not answer_list:
            answer_list = [None]
        for answer in answer_list:
            obs_dns.append(
                DNSObservation.from_measurement(
                    msmt, query, answer, idx, fingerprintdb, netinfodb, target
                )
            )
            idx += 1

    return obs_dns


@dataclass
class TCPObservation(Observation):
    __table_name__ = "obs_tcp"

    domain_name: str

    ip: str
    port: int

    failure: Failure

    ip_asn: Optional[int] = None
    ip_as_org_name: Optional[str] = None
    ip_as_cc: Optional[str] = None
    ip_cc: Optional[str] = None

    transaction_id: Optional[int] = None

    @staticmethod
    def from_measurement(
        msmt: BaseMeasurement,
        res: TCPConnect,
        idx: int,
        ip_to_domain: Dict[str, str],
        netinfodb: NetinfoDB,
        target: str,
    ) -> "TCPObservation":
        tcpo = TCPObservation(
            observation_id=f"{msmt.measurement_uid}_tcp_{idx}",
            timestamp=make_timestamp(msmt, res.t),
            ip=res.ip,
            port=res.port,
            failure=normalize_failure(res.status.failure),
            domain_name=ip_to_domain.get(res.ip, ""),
            target=target,
            transaction_id=res.transaction_id,
            **make_base_observation_meta(msmt, netinfodb),
        )

        ip_info = netinfodb.lookup_ip(tcpo.timestamp, res.ip)
        if ip_info:
            tcpo.ip_asn = ip_info.as_info.asn
            tcpo.ip_as_org_name = ip_info.as_info.as_org_name
            tcpo.ip_as_cc = ip_info.as_info.as_cc

            tcpo.ip_cc = ip_info.cc

        return tcpo


def make_tcp_observations(
    msmt: BaseMeasurement,
    tcp_connect: Optional[List[TCPConnect]],
    netinfodb: NetinfoDB,
    ip_to_domain: Dict[str, str] = {},
    target: str = "",
) -> List[TCPObservation]:
    obs_tcp = []
    if not tcp_connect:
        return obs_tcp

    for idx, res in enumerate(tcp_connect):
        obs_tcp.append(
            TCPObservation.from_measurement(
                msmt, res, idx, ip_to_domain, netinfodb, target
            )
        )
    return obs_tcp


def network_events_until_connect(
    network_events: List[NetworkEvent],
) -> List[NetworkEvent]:
    ne_list = []
    for ne in network_events:
        if ne.operation == "connect":
            break
        ne_list.append(ne)
    return ne_list


def find_tls_handshake_network_events(
    tls_handshake: TLSHandshake, network_events: Optional[List[NetworkEvent]]
) -> Optional[List[NetworkEvent]]:
    if not network_events:
        return None

    current_event_window = []
    for idx, ne in enumerate(network_events):
        if ne.operation == "connect":
            current_event_window = []
        current_event_window.append(ne)
        # We identify the network_event for the given TLS handshake based on the
        # fact that the timestamp on tls_handshake_done event is the same as the
        # tls_handshake time
        if ne.operation == "tls_handshake_done" and ne.t == tls_handshake.t:
            current_event_window += network_events_until_connect(network_events[idx:])
            return current_event_window
    return None


@dataclass
class TLSObservation(Observation):
    __table_name__ = "obs_tls"

    domain_name: str

    failure: Failure

    server_name: str
    tls_version: str
    cipher_suite: str

    ip: Optional[str] = None
    port: Optional[int] = None

    ip_asn: Optional[int] = None
    ip_as_org_name: Optional[str] = None
    ip_as_cc: Optional[str] = None
    ip_cc: Optional[str] = None

    is_certificate_valid: Optional[bool] = None

    end_entity_certificate_fingerprint: Optional[str] = None
    end_entity_certificate_subject: Optional[str] = None
    end_entity_certificate_subject_common_name: Optional[str] = None
    end_entity_certificate_issuer: Optional[str] = None
    end_entity_certificate_issuer_common_name: Optional[str] = None
    end_entity_certificate_san_list: List[str] = field(default_factory=list)
    end_entity_certificate_not_valid_after: Optional[datetime] = None
    end_entity_certificate_not_valid_before: Optional[datetime] = None
    peer_certificates: List[bytes] = field(default_factory=list)
    certificate_chain_length: Optional[int] = None

    tls_handshake_read_count: Optional[int] = None
    tls_handshake_write_count: Optional[int] = None
    tls_handshake_read_bytes: Optional[float] = None
    tls_handshake_write_bytes: Optional[float] = None
    tls_handshake_last_operation: Optional[str] = None
    tls_handshake_time: Optional[float] = None

    transaction_id: Optional[int] = None

    @staticmethod
    def from_measurement(
        msmt: BaseMeasurement,
        tls_h: TLSHandshake,
        network_events: Optional[List[NetworkEvent]],
        idx: int,
        ip_to_domain: Dict[str, str],
        netinfodb: NetinfoDB,
        cert_store: Optional[TLSCertStore] = None,
        validate_domain: Callable[[str, str, List[str]], bool] = lambda x, y, z: True,
        target: str = "",
    ) -> "TLSObservation":
        tlso = TLSObservation(
            observation_id=f"{msmt.measurement_uid}_tls_{idx}",
            timestamp=make_timestamp(msmt, tls_h.t),
            server_name=tls_h.server_name if tls_h.server_name else "",
            domain_name=tls_h.server_name if tls_h.server_name else "",
            tls_version=tls_h.tls_version if tls_h.tls_version else "",
            cipher_suite=tls_h.cipher_suite if tls_h.cipher_suite else "",
            end_entity_certificate_san_list=[],
            failure=normalize_failure(tls_h.failure),
            target=target,
            transaction_id=tls_h.transaction_id,
            **make_base_observation_meta(msmt, netinfodb),
        )

        if tls_h.address:
            p = urlsplit("//" + tls_h.address)
            tlso.ip = p.hostname
            tlso.port = p.port

        tls_network_events = find_tls_handshake_network_events(tls_h, network_events)
        if tls_network_events:
            if tls_network_events[0].address:
                p = urlsplit("//" + tls_network_events[0].address)
                tlso.ip = p.hostname
                tlso.port = p.port
                if tlso.ip and tlso.ip in ip_to_domain:
                    tlso.domain_name = ip_to_domain[tlso.ip]

            tlso.tls_handshake_time = tls_network_events[-1].t - tls_network_events[0].t
            tlso.tls_handshake_read_count = 0
            tlso.tls_handshake_write_count = 0
            tlso.tls_handshake_read_bytes = 0
            tlso.tls_handshake_write_bytes = 0
            for ne in tls_network_events:
                if ne.operation == "write":
                    if ne.num_bytes:
                        tlso.tls_handshake_write_count += 1
                        tlso.tls_handshake_write_bytes += ne.num_bytes
                    tlso.tls_handshake_last_operation = (
                        f"write_{tlso.tls_handshake_write_count}"
                    )
                elif ne.operation == "read" and ne.num_bytes:
                    if ne.num_bytes:
                        tlso.tls_handshake_read_count += 1
                        tlso.tls_handshake_read_bytes += ne.num_bytes
                    tlso.tls_handshake_last_operation = (
                        f"read_{tlso.tls_handshake_read_count}"
                    )

        if tls_h.peer_certificates:
            try:
                tlso.peer_certificates = list(
                    map(lambda c: b64decode(c.data), tls_h.peer_certificates)
                )
            except Exception:
                log.error("failed to decode peer_certificates")

            tlso.certificate_chain_length = len(tls_h.peer_certificates)
            try:
                cert_meta = get_certificate_meta(tls_h.peer_certificates[0])
                tlso.end_entity_certificate_fingerprint = cert_meta.fingerprint
                tlso.end_entity_certificate_subject = cert_meta.subject
                tlso.end_entity_certificate_subject_common_name = (
                    cert_meta.subject_common_name
                )
                tlso.end_entity_certificate_issuer = cert_meta.issuer
                tlso.end_entity_certificate_issuer_common_name = (
                    cert_meta.issuer_common_name
                )
                tlso.end_entity_certificate_not_valid_after = cert_meta.not_valid_after
                tlso.end_entity_certificate_not_valid_before = (
                    cert_meta.not_valid_before
                )
                tlso.end_entity_certificate_san_list = cert_meta.san_list
            except Exception as exc:
                log.error(exc)
                log.error(
                    f"Failed to extract certificate meta for {msmt.measurement_uid}"
                )

        if cert_store and tlso.peer_certificates:
            try:
                cn, san_list = cert_store.validate_cert_chain(
                    tlso.timestamp, tlso.peer_certificates
                )
                tlso.is_certificate_valid = validate_domain(
                    tlso.server_name, cn, san_list
                )
            except InvalidCertificateChain:
                tlso.is_certificate_valid = False

        elif tls_h.no_tls_verify == False:
            if tlso.failure in (
                "ssl_invalid_hostname",
                "ssl_unknown_authority",
                "ssl_invalid_certificate",
            ):
                tlso.is_certificate_valid = False
            elif not tlso.failure:
                tlso.is_certificate_valid = True

        return tlso


def make_tls_observations(
    msmt: BaseMeasurement,
    tls_handshakes: Optional[List[TLSHandshake]],
    network_events: Optional[List[NetworkEvent]],
    netinfodb: NetinfoDB,
    ip_to_domain: Dict[str, str] = {},
    cert_store: Optional[TLSCertStore] = None,
    validate_domain: Callable[[str, str, List[str]], bool] = lambda x, y, z: True,
    target: str = "",
) -> List[TLSObservation]:
    obs_tls = []
    if not tls_handshakes:
        return obs_tls

    for idx, tls_h in enumerate(tls_handshakes):
        obs_tls.append(
            TLSObservation.from_measurement(
                msmt,
                tls_h,
                network_events,
                idx,
                ip_to_domain,
                netinfodb,
                cert_store,
                validate_domain,
                target,
            )
        )
    return obs_tls


def make_ip_to_domain(dns_observations: List[DNSObservation]) -> Dict[str, str]:
    ip_to_domain = {}
    for obs in dns_observations:
        # TODO: do we want to filter out also CNAMEs?
        if not obs.answer:
            continue
        # TODO: this is only really valid for web_connectivity and even there it
        # only works if there isn't a redirect chain with domains that map to
        # different domains.
        # What we should do is make this into a list and then figure out which
        # is the relevant domain for a particular resolution by looking at other data.
        # Better yet, this should be marked inside of the measurement itself.
        # TODO(sbs): what happens in the engine if I encounter in a redirect chain something like:
        # https://example.com/ -> https://www.example.com/ where both
        # www.example.com and example.com map to the same IP 1.2.3.4?
        # Will we record perform two tcp_connect measurements or only one?
        # If it's two we need to tell which one is pertaining to one or the other.
        # If it's one, then we need to make changes in the base dataformat so
        # that we can express that a single tcp_connect experiment is pertaining
        # to two different domains.
        if obs.answer in ip_to_domain:
            log.error(
                f"multiple resolutions for the same IP {obs.answer}:{obs.domain_name}"
            )
        ip_to_domain[obs.answer] = obs.domain_name
    return ip_to_domain


def make_web_connectivity_observations(
    msmt: WebConnectivity,
    fingerprintdb: FingerprintDB,
    netinfodb: NetinfoDB,
    target: str = "",
) -> Tuple[
    List[DNSObservation],
    List[TCPObservation],
    List[TLSObservation],
    List[HTTPObservation],
]:
    http_observations = make_http_observations(
        msmt, msmt.test_keys.requests, fingerprintdb, netinfodb, target
    )

    dns_observations = make_dns_observations(
        msmt, msmt.test_keys.queries, fingerprintdb, netinfodb
    )
    ip_to_domain = make_ip_to_domain(dns_observations)
    tcp_observations = make_tcp_observations(
        msmt, msmt.test_keys.tcp_connect, netinfodb, ip_to_domain
    )

    tls_observations = make_tls_observations(
        msmt,
        msmt.test_keys.tls_handshakes,
        msmt.test_keys.network_events,
        netinfodb,
        ip_to_domain,
    )

    # Here we take dns measurements and compare them to what we see in the tls
    # data and check for TLS consistency.
    tls_valid_ip_to_domain = {}
    for obs in filter(
        lambda o: o.ip and o.domain_name,
        tls_observations,
    ):
        tls_valid_ip_to_domain[obs.ip] = tls_valid_ip_to_domain.get(obs.ip, {})
        tls_valid_ip_to_domain[obs.ip][obs.domain_name] = obs.is_certificate_valid
    enriched_dns_observations = []
    for dns_obs in dns_observations:
        if dns_obs.answer:
            valid_domains = tls_valid_ip_to_domain.get(dns_obs.answer, {})
            dns_obs.is_tls_consistent = valid_domains.get(dns_obs.domain_name, None)
        enriched_dns_observations.append(dns_obs)

    return (
        enriched_dns_observations,
        tcp_observations,
        tls_observations,
        http_observations,
    )


def make_signal_observations(
    msmt: Signal, fingerprintdb: FingerprintDB, netinfodb: NetinfoDB
) -> Tuple[
    List[DNSObservation],
    List[TCPObservation],
    List[TLSObservation],
    List[HTTPObservation],
]:
    cert_store = TLSCertStore(SIGNAL_PEM_STORE)
    http_observations = make_http_observations(
        msmt, msmt.test_keys.requests, fingerprintdb, netinfodb
    )

    dns_observations = make_dns_observations(
        msmt, msmt.test_keys.queries, fingerprintdb, netinfodb
    )

    ip_to_domain = make_ip_to_domain(dns_observations)
    tcp_observations = make_tcp_observations(
        msmt, msmt.test_keys.tcp_connect, netinfodb, ip_to_domain
    )

    tls_observations = list(
        make_tls_observations(
            msmt,
            msmt.test_keys.tls_handshakes,
            msmt.test_keys.network_events,
            netinfodb,
            ip_to_domain,
            cert_store,
        )
    )

    # Because we are using certificate pinning for Signal, if we got a
    # successful TLS handshake with some endpoint, it's bound to be TLS
    # consistent.
    tls_consistency_map = {}
    for tls_obs in tls_observations:
        if tls_obs.ip and tls_obs.domain_name:
            tls_consistency_map[tls_obs.ip] = tls_obs.is_certificate_valid

    enriched_dns_observations = []
    for dns_obs in dns_observations:
        if dns_obs.answer:
            dns_obs.is_tls_consistent = tls_consistency_map.get(dns_obs.answer, None)
        enriched_dns_observations.append(dns_obs)

    return (
        enriched_dns_observations,
        tcp_observations,
        tls_observations,
        http_observations,
    )


def make_dnscheck_observations(
    msmt: DNSCheck,
    fingerprintdb: FingerprintDB,
    netinfodb: NetinfoDB,
) -> Tuple[
    List[DNSObservation],
    List[TCPObservation],
    List[TLSObservation],
    List[HTTPObservation],
]:
    dns_observations = []
    http_observations = []
    tcp_observations = []
    tls_observations = []

    ip_to_domain = {}
    if msmt.test_keys.bootstrap:
        dns_observations += make_dns_observations(
            msmt, msmt.test_keys.bootstrap.queries, fingerprintdb, netinfodb
        )
        ip_to_domain = make_ip_to_domain(dns_observations)

    lookup_map = msmt.test_keys.lookups or {}
    for lookup in lookup_map.values():
        dns_observations += make_dns_observations(
            msmt, lookup.queries, fingerprintdb, netinfodb
        )

        http_observations = make_http_observations(
            msmt, lookup.requests, fingerprintdb, netinfodb
        )

        tcp_observations += make_tcp_observations(
            msmt, lookup.tcp_connect, netinfodb, ip_to_domain
        )

        tls_observations += make_tls_observations(
            msmt, lookup.tls_handshakes, lookup.network_events, netinfodb, ip_to_domain
        )

    return dns_observations, tcp_observations, tls_observations, http_observations


def make_tor_observations(
    msmt: Tor,
    fingerprintdb: FingerprintDB,
    netinfodb: NetinfoDB,
) -> Tuple[
    List[DNSObservation],
    List[TCPObservation],
    List[TLSObservation],
    List[HTTPObservation],
]:
    dns_observations = []
    http_observations = []
    tcp_observations = []
    tls_observations = []

    ip_to_domain = {}
    for target_id, target_msmt in msmt.test_keys.targets.items():
        http_observations += make_http_observations(
            msmt, target_msmt.requests, fingerprintdb, netinfodb, target=target_id
        )
        dns_observations += make_dns_observations(
            msmt, target_msmt.queries, fingerprintdb, netinfodb, target=target_id
        )
        tcp_observations += make_tcp_observations(
            msmt, target_msmt.tcp_connect, netinfodb, ip_to_domain, target=target_id
        )
        tls_observations += make_tls_observations(
            msmt,
            target_msmt.tls_handshakes,
            target_msmt.network_events,
            netinfodb,
            ip_to_domain,
        )

    return dns_observations, tcp_observations, tls_observations, http_observations



import json
import shutil
import gzip
import logging
import hashlib

from typing import List
from pathlib import Path
from datetime import datetime, date
from collections import OrderedDict, namedtuple
from dataclasses import dataclass

import xml.etree.ElementTree as ET

import requests
from requests.adapters import HTTPAdapter, Retry

import maxminddb

from oonidata.datautils import is_ip_bogon

log = logging.getLogger("oonidata.processing")


retry_strategy = Retry(total=4, backoff_factor=0.1)

req_session = requests.Session()
req_session.mount("http://", HTTPAdapter(max_retries=retry_strategy))
req_session.mount("https://", HTTPAdapter(max_retries=retry_strategy))


def file_sha1_hexdigest(filepath: Path):
    h = hashlib.sha1()
    with filepath.open("rb") as in_file:
        while True:
            b = in_file.read(2**16)
            if not b:
                break
            h.update(b)
    return h.hexdigest()


IAItem = namedtuple("IAItem", ["identifier", "filename", "sha1"])


def list_all_ia_items(identifier: str) -> List[IAItem]:
    ia_items = []
    resp = req_session.get(
        f"https://archive.org/download/{identifier}/{identifier}_files.xml"
    )
    if resp.status_code == 404:
        return []

    resp.raise_for_status()
    tree = ET.fromstring(resp.text)
    for f in tree:
        fname = f.get("name")
        if not fname:
            continue

        sha1 = f.find("sha1")
        if sha1 is not None:
            sha1 = sha1.text
        ia_items.append(IAItem(identifier=identifier, filename=fname, sha1=sha1))

    return ia_items


def download_ia_item(ia_item: IAItem, output_path: Path):
    url = f"https://archive.org/download/{ia_item.identifier}/{ia_item.filename}"
    with req_session.get(url, stream=True) as resp:
        resp.raise_for_status()
        with output_path.with_suffix(".tmp").open("wb") as out_file:
            for b in resp.iter_content(chunk_size=2**16):
                out_file.write(b)

    output_path.with_suffix(".tmp").rename(output_path)
    file_sha1 = file_sha1_hexdigest(output_path)
    assert file_sha1 == ia_item.sha1, f"{file_sha1} != {ia_item.sha1}"


@dataclass
class ASInfo:
    asn: int

    as_org_name: str
    as_cc: str
    as_name: str


@dataclass
class IPInfo:
    as_info: ASInfo
    cc: str


class NetinfoDB:
    def __init__(
        self,
        datadir: Path = Path("datadir"),
        download: bool = False,
    ):
        self.datadir = datadir
        self.ip2country_as_dir = self.datadir / "ip2country-as"
        if download:
            self.refresh_netinfodb()

        try:
            with (self.ip2country_as_dir / "all_as_org_map.json").open() as in_file:
                self.as_org_map = json.load(in_file)
        except FileNotFoundError:
            log.error("unable to find all_as_org_map.json. Try setting download = True")
            raise

        self.load_ip2country_as()
        self.readers = {}

    def refresh_netinfodb(self):
        self.ip2country_as_dir.mkdir(parents=True, exist_ok=True)

        for item in list_all_ia_items("ip2country-as"):
            if (
                not item.filename.endswith(".mmdb.gz")
                and not item.filename == "all_as_org_map.json"
            ):
                continue

            output_path = self.ip2country_as_dir / item.filename
            if output_path.exists() and file_sha1_hexdigest(output_path) == item.sha1:
                continue

            log.info(f"downloading {item.filename}")
            download_ia_item(ia_item=item, output_path=output_path)

            if output_path.name.endswith(".gz"):
                dst_path = output_path.with_suffix("")
                log.info(f"decompressing to {dst_path}")
                with gzip.open(output_path) as in_file, dst_path.with_suffix(
                    ".tmp"
                ).open("wb") as out_file:
                    shutil.copyfileobj(in_file, out_file)
                dst_path.with_suffix(".tmp").rename(dst_path)

    def get_reader(self, db_path: Path):
        if db_path in self.readers:
            return self.readers[db_path]
        self.readers[db_path] = maxminddb.open_database(str(db_path))
        return self.readers[db_path]

    def load_ip2country_as(self):
        """
        Populate the OrderedDict of decompressed geoip database files that are
        inside of the datadir.
        We expect there to be two files on each date and the format of the file
        is: "%Y%m%d-ip2country_as.mmdb" ex. 20181009-ip2country_as.mmdb.gz
        """
        self.databases = OrderedDict()
        for db_path in sorted(self.ip2country_as_dir.glob("*.mmdb")):
            ts = datetime.strptime(db_path.name.split("-")[0], "%Y%m%d").date()
            self.databases[ts] = db_path

        assert len(self.databases) > 0, "Did not find any geoip database files"

    def find_db_for_date(self, day: date):
        """
        Find DB for date will return a dictionary with asn and country keys set
        which is closest in time and <= day.
        """
        chosen_db = list(self.databases.values())[0]
        for ts, db in self.databases.items():
            if ts > day:
                break
            chosen_db = db
        return chosen_db

    def lookup_asn(self, day: datetime, asn: int) -> ASInfo:
        """
        Returns information about a particular ASN on a given day, if known.
        """
        day_str = day.strftime("%Y%m%d")
        org_name, name, country = ("", "", "")
        try:
            meta_list = self.as_org_map[str(asn)]
            org_name, country, name = meta_list[0][:3]
            for meta in meta_list:
                if meta[2] > day_str:
                    break
                org_name, country, name = meta[:3]
        except KeyError:
            log.error(f"Unable to locate ASN {asn}")

        return ASInfo(asn=asn, as_org_name=org_name, as_cc=country, as_name=name)

    def lookup_ip(self, day: datetime, ip: str) -> IPInfo:
        unknown_ipinfo = IPInfo(
            ASInfo(
                asn=0,
                as_org_name="",
                as_name="",
                as_cc="",
            ),
            cc="ZZ",
        )

        db_path = self.find_db_for_date(day.date())
        assert db_path is not None

        try:
            if is_ip_bogon(ip):
                return IPInfo(
                    ASInfo(
                        asn=64666,
                        as_org_name="Bogon",
                        as_cc="ZZ",
                        as_name="",
                    ),
                    cc="ZZ",
                )
        except ValueError:
            return unknown_ipinfo

        reader = self.get_reader(db_path)
        res = None
        try:
            res = reader.get(ip)
        except ValueError:
            pass

        if not res:
            log.error(f"Failed to lookup {ip}")
            return unknown_ipinfo

        return IPInfo(
            ASInfo(
                asn=res.get("autonomous_system_number", 0),
                as_org_name=res.get("autonomous_system_organization", ""),
                as_cc=res.get("autonomous_system_country", ""),
                as_name=res.get("autonomous_system_name", ""),
            ),
            cc=res.get("country", {}).get("iso_code", "ZZ"),
        )



import time
import logging
import traceback
import orjson

from tqdm import tqdm

from datetime import date, timedelta
from dataclasses import asdict, fields

from typing import (
    Tuple,
    List,
    Generator,
    Type,
    TypeVar,
    Optional,
    Union,
    Iterable,
    Dict,
)

from oonidata.datautils import one_day_dict, is_ip_bogon
from oonidata.observations import (
    NettestObservation,
    DNSObservation,
    Observation,
    make_tor_observations,
    make_signal_observations,
    make_web_connectivity_observations,
    make_dnscheck_observations,
)
from oonidata.dataformat import DNSCheck, load_measurement
from oonidata.dataformat import BaseMeasurement, WebConnectivity, Tor
from oonidata.fingerprintdb import FingerprintDB
from oonidata.netinfo import NetinfoDB

from oonidata.dataclient import (
    MeasurementListProgress,
    iter_measurements,
    ProgressStatus,
)
from oonidata.db.connections import (
    DatabaseConnection,
    ClickhouseConnection,
    CSVConnection,
)

log = logging.getLogger("oonidata.processing")


def observation_field_names(obs_class: Type[Observation]) -> List[str]:
    return list(map(lambda dc: dc.name, fields(obs_class)))


def make_observation_row(observation: Observation) -> dict:
    return asdict(observation)


def write_observations_to_db(
    db: DatabaseConnection, observations: Iterable[Observation]
) -> None:
    for obs in observations:
        row = make_observation_row(obs)
        db.write_row(obs.__table_name__, row)


def domains_in_a_day(
    day: date, db: ClickhouseConnection, probe_cc: Optional[str]
) -> List[str]:
    q = """SELECT DISTINCT(domain_name) FROM obs_dns
    WHERE timestamp >= %(start_day)s
    AND timestamp <= %(end_day)s
    """
    params = one_day_dict(day)
    if probe_cc:
        q += "AND probe_cc = %(probe_cc)s"
        params["probe_cc"] = probe_cc
    res = db.execute(q, params)
    assert isinstance(res, list)
    return [row[0] for row in res]


def dns_observations_by_session(
    day: date,
    domain_name: str,
    db: ClickhouseConnection,
    probe_cc: Optional[str] = None,
) -> Generator[List[DNSObservation], None, None]:
    # I wish I had an ORM...
    field_names = observation_field_names(DNSObservation)

    q_params = one_day_dict(day)
    q_params["domain_name"] = domain_name

    q = "SELECT "
    q += ",\n".join(field_names)
    q += """
    FROM obs_dns
    WHERE domain_name = %(domain_name)s
    AND timestamp >= %(start_day)s
    AND timestamp <= %(end_day)s
    """
    if probe_cc:
        q += "AND probe_cc = %(probe_cc)s\n"
        q_params["probe_cc"] = probe_cc

    q += "ORDER BY report_id, measurement_uid;"

    # Put all the DNS observations from the same testing session into a list and yield it
    dns_obs_session = []
    last_obs_session_id = None
    res = db.execute(q, q_params)
    assert isinstance(res, list)
    for rows in res:
        obs_dict = {field_names[idx]: val for idx, val in enumerate(rows)}
        dns_obs = DNSObservation(**obs_dict)

        if last_obs_session_id and last_obs_session_id != dns_obs.report_id:
            yield dns_obs_session
            dns_obs_session = [dns_obs]
            last_obs_session_id = dns_obs.report_id
        else:
            dns_obs_session.append(dns_obs)
    if len(dns_obs_session) > 0:
        yield dns_obs_session


T = TypeVar("T", bound="Observation")


def observations_in_session(
    day: date,
    domain_name: str,
    obs_class: Type[T],
    report_id: str,
    db: ClickhouseConnection,
) -> List[T]:
    observation_list = []
    field_names = observation_field_names(obs_class)
    q = "SELECT "
    q += ",\n".join(field_names)
    q += " FROM " + obs_class.__table_name__
    q += """
    WHERE domain_name = %(domain_name)s
    AND timestamp >= %(start_day)s
    AND timestamp <= %(end_day)s
    AND report_id = %(report_id)s;
    """
    q_params = one_day_dict(day)
    q_params["domain_name"] = domain_name
    q_params["report_id"] = report_id

    res = db.execute(q, q_params)
    assert isinstance(res, list)
    for rows in res:
        obs_dict = {field_names[idx]: val for idx, val in enumerate(rows)}
        observation_list.append(obs_class(**obs_dict))
    return observation_list


def nettest_processor(
    msmt: DNSCheck,
    db: DatabaseConnection,
    fingerprintdb: FingerprintDB,
    netinfodb: NetinfoDB,
) -> None:
    for obs_group in make_dnscheck_observations(msmt, fingerprintdb, netinfodb):
        write_observations_to_db(db, obs_group)


def generic_processor(
    msmt: BaseMeasurement,
    db: DatabaseConnection,
    netinfodb: NetinfoDB,
) -> None:
    write_observations_to_db(db, [NettestObservation.from_measurement(msmt, netinfodb)])


nettest_make_obs_map = {
    "web_connectivity": make_web_connectivity_observations,
    "dnscheck": make_dnscheck_observations,
    "tor": make_tor_observations,
    "signal": make_signal_observations,
}


def process_msmt_dict(
    msmt_dict: Dict,
    db: Union[ClickhouseConnection, CSVConnection],
    netinfodb: NetinfoDB,
    fingerprintdb: FingerprintDB,
):
    msmt = load_measurement(msmt_dict)
    generic_processor(msmt, db, netinfodb)

    if msmt.test_name in nettest_make_obs_map:
        for obs_group in nettest_make_obs_map[msmt.test_name](
            msmt, fingerprintdb, netinfodb
        ):
            write_observations_to_db(db, obs_group)


def process_day(
    db: Union[ClickhouseConnection, CSVConnection],
    fingerprintdb: FingerprintDB,
    netinfodb: NetinfoDB,
    day: date,
    test_name=[],
    probe_cc=[],
    start_at_idx=0,
    fast_fail=False,
) -> Tuple[float, date]:
    t0 = time.monotonic()
    with tqdm(unit="B", unit_scale=True) as pbar:

        def progress_callback(p: MeasurementListProgress):
            if p.progress_status == ProgressStatus.LISTING:
                if not pbar.total:
                    pbar.total = p.total_prefixes
                pbar.update(1)
                pbar.set_description(
                    f"listed {p.total_file_entries} files in {p.current_prefix_idx}/{p.total_prefixes} prefixes"
                )
                return

            if p.progress_status == ProgressStatus.DOWNLOAD_BEGIN:
                pbar.unit = "B"
                pbar.reset(total=p.total_file_entry_bytes)

            pbar.set_description(
                f"downloading {p.current_file_entry_idx}/{p.total_file_entries} files"
            )
            pbar.update(p.current_file_entry_bytes)

        for idx, msmt_dict in enumerate(
            iter_measurements(
                probe_cc=probe_cc,
                test_name=test_name,
                start_day=day,
                end_day=day + timedelta(days=1),
                progress_callback=progress_callback,
            )
        ):
            pbar.set_description(f"idx {idx}")
            if idx < start_at_idx:
                continue
            try:
                process_msmt_dict(
                    msmt_dict=msmt_dict,
                    db=db,
                    netinfodb=netinfodb,
                    fingerprintdb=fingerprintdb,
                )
            except Exception as exc:
                log.error(f"failed at idx:{idx} {exc}")
                with open("bad_msmts.jsonl", "ab+") as out_file:
                    out_file.write(orjson.dumps(msmt_dict))
                    out_file.write(b"\n")
                with open("bad_msmts_fail_log.txt", "a+") as out_file:
                    out_file.write(traceback.format_exc())
                    out_file.write("ENDTB----\n")
                if fast_fail:
                    raise exc

    return time.monotonic() - t0, day



#
# Normalize YAML reports
#

from datetime import datetime
from itertools import groupby
import functools
import hashlib
import logging
import re
import string
import uuid

import yaml
import yaml.parser

from oonidata.dataformat import trivial_id

log = logging.getLogger("normalize")


class UnsupportedTestError(Exception):
    pass


test_name_mappings = {
    "http_host": "http_host",
    "HTTP Host": "http_host",
    "http_requests_test": "http_requests",
    "http_requests": "http_requests",
    "HTTP Requests Test": "http_requests",
    "bridge_reachability": "bridge_reachability",
    "bridgereachability": "bridge_reachability",
    "TCP Connect": "tcp_connect",
    "tcp_connect": "tcp_connect",
    "DNS tamper": "dns_consistency",
    "dnstamper": "dns_consistency",
    "dns_consistency": "dns_consistency",
    "HTTP Invalid Request Line": "http_invalid_request_line",
    "http_invalid_request_line": "http_invalid_request_line",
    "http_header_field_manipulation": "http_header_field_manipulation",
    "HTTP Header Field Manipulation": "http_header_field_manipulation",
    "Multi Protocol Traceroute Test": "multi_protocol_traceroute",
    "multi_protocol_traceroute_test": "multi_protocol_traceroute",
    "multi_protocol_traceroute": "multi_protocol_traceroute",
    "traceroute": "multi_protocol_traceroute",
    "parasitic_traceroute_test": "parasitic_traceroute",
    "parasitic_tcp_traceroute_test": "parasitic_traceroute",
    "tls-handshake": "tls_handshake",
    "tls_handshake": "tls_handshake",
    "dns_injection": "dns_injection",
    "captivep": "captive_portal",
    "captiveportal": "captive_portal",
    "HTTPFilteringBypass": "http_filtering_bypass",
    "httpfilteringbypass": "http_filtering_bypass",
    "HTTPTrix": "http_trix",
    "httptrix": "http_trix",
    "http_test": "http_test",
    "http_url_list": "http_url_list",
    "dns_spoof": "dns_spoof",
    "netalyzrwrapper": "netalyzr_wrapper",
    "meek_fronted_requests_test": "meek_fronted_requests_test",
    "lantern_circumvention_tool_test": "lantern_circumvention_tool_test",
    "psiphon_test": "psiphon_test",
    "this_test_is_nameless": "this_test_is_nameless",
    "test_get_random_capitalization": "http_header_field_manipulation",
    "test_put_random_capitalization": "http_header_field_manipulation",
    "test_post_random_capitalization": "http_header_field_manipulation",
    "test_random_big_request_method": "http_invalid_request_line",
    "test_random_invalid_field_count": "http_invalid_request_line",
    "summary": "invalid",
    "test_get": "invalid",
    "test_post": "invalid",
    "test_put": "invalid",
    "test_send_host_header": "invalid",
}

schema = [
    "id",
    "input",
    "input_hashes",
    "report_id",
    "report_filename",
    "options",
    "probe_cc",
    "probe_asn",
    "probe_ip",
    "probe_city",
    "backend_version",
    "data_format_version",
    "test_name",
    "test_version",
    "test_start_time",
    "measurement_start_time",
    "test_runtime",
    "test_helpers",
    "software_name",
    "software_version",
    "bucket_date",
    "test_keys",
]

# Some values overlap across categories e.g. captive_portal
test_categories = {
    "dnst": {"dns_consistency", "dns_injection", "captive_portal"},
    "process": {"lantern", "psiphon"},
    "httpt": {
        "http_requests",
        "meek_fronted_requests",
        "domclass_collector",
        "http_keyword_filtering",
        "http_uk_mobile_networks",
        "http_header_field_manipulation",
        "http_url_list",
        "http_host",
        "squid",
        "captive_portal",
        "psiphon",
        "tor_http_requests_test",
    },
    "scapyt": {
        "chinatrigger",
        "keyword_filtering",
        "parasitic_traceroute",
        "multi_protocol_traceroute",
        "dns_spoof",
    },
    "tcpt": {"http_filtering_bypass", "http_invalid_request_line", "http_trix"},
}

regexps = dict(
    ipv4=r"(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})",
    hostname=(r"([a-zA-Z0-9](?:(?:[a-zA-Z0-9-]*|(?<!-)\.(?![-.]))*[a-zA-Z0-9]+)?)"),
)


## Simhash generation

simhash_re = re.compile(r"[\w\u4e00-\u9fcc]+")


def gen_simhash(s):
    content = s.lower()
    content = "".join(re.findall(simhash_re, content))
    mx = max(len(content) - 4 + 1, 1)
    features = [content[i : i + 4] for i in range(mx)]
    features = ((k, sum(1 for _ in g)) for k, g in groupby(sorted(features)))
    v = [0] * 64
    masks = [1 << i for i in range(64)]
    for h, w in features:
        h = h.encode("utf-8")
        h = int(hashlib.md5(h).hexdigest(), 16)
        for i in range(64):
            v[i] += w if h & masks[i] else -w
    ans = 0
    for i in range(64):
        if v[i] > 0:
            ans |= masks[i]
    return ans


### Normalize entries across format versions ###


def nest_test_keys(entry):
    if entry["test_keys"] is None:
        entry["test_keys"] = {}
    for test_key in set(entry.keys()) - set(schema):
        entry["test_keys"][test_key] = entry.pop(test_key)

    return entry


def normalize_str(body):
    if body is None:
        return None
    assert len(body) < 10000
    if isinstance(body, bytes):
        return body.decode("UTF-8", "backslashreplace")
    return body.replace("\0", "")


def normalize_body(body):
    if body is None:
        return None

    if isinstance(body, bytes):
        return body.decode("UTF-8", "backslashreplace")

    return body.replace("\0", "")


def match(pattern, source):
    found = re.search(pattern, source)
    if found:
        return found.group(1)
    return ""


def normalize_httpt(entry):
    def normalize_headers(headers):
        # XXX: data loss -- ordering, formatting, duplicate headers, whitespace
        normalized_headers = {}
        for name, values in headers:
            value = values[0]
            if isinstance(value, list):
                value = value[0]
            normalized_headers[name] = normalize_str(value)
        return normalized_headers

    experiment_requests = []
    control_requests = []

    url_option_idx = None
    url_option_names = ["--url", "-u"]
    for url_option in url_option_names:
        try:
            url_option_idx = entry.get("options").index(url_option) + 1
        except (ValueError, AttributeError):
            continue

    if url_option_idx is not None and entry["input"] is None:
        entry["input"] = entry["options"][url_option_idx]

    # This is needed to fix the requests and responses in the
    # tor_http_requests test.
    if entry["test_keys"].get("request", None) and entry["test_keys"].get(
        "response", None
    ):
        entry["test_keys"]["requests"] = entry["test_keys"].get("requests", [])
        entry["test_keys"]["requests"].append(
            {
                "response": entry["test_keys"].pop("response"),
                "request": entry["test_keys"].pop("request"),
            }
        )

    for session in entry["test_keys"].get("requests", []):
        if isinstance(session.get("response"), dict):
            session["response"]["body"] = normalize_body(session["response"]["body"])
            session["response"]["headers"] = normalize_headers(
                session["response"]["headers"]
            )
        else:
            session["response"] = {"body": None, "headers": {}}

        if isinstance(session.get("request"), dict):
            session["request"]["body"] = normalize_body(session["request"]["body"])
            session["request"]["headers"] = normalize_headers(
                session["request"]["headers"]
            )
        else:
            session["request"] = {"body": None, "headers": {}}

        is_tor = False
        exit_ip = None
        exit_name = None
        if session["request"]["url"].startswith("shttp"):
            session["request"]["url"] = session["request"]["url"].replace(
                "shttp://", "http://"
            )
            is_tor = True
        elif session["request"].get("tor") is True:
            is_tor = True
        elif session["request"].get("tor") in [False, None, {"is_tor": False}]:
            is_tor = False
        elif session["request"].get("tor", {}).get("is_tor") is True:
            is_tor = True
            exit_ip = session["request"].get("tor", {}).get("exit_ip", None)
            exit_name = session["request"].get("tor", {}).get("exit_name", None)
        else:
            log.error("Could not detect tor or not tor status")
            log.debug(session)

        session["request"]["tor"] = {
            "is_tor": is_tor,
            "exit_ip": exit_ip,
            "exit_name": exit_name,
        }
        session["response_length"] = None
        for k, v in session["response"]["headers"].items():
            # sort of re-normalisation from body back to binary
            if k.lower() == "content-length":
                session["response_length"] = v
        if is_tor is True:
            control_requests.append(session)
        else:
            experiment_requests.append(session)
    entry["test_keys"]["requests"] = []
    try:
        entry["test_keys"]["requests"].append(experiment_requests.pop(0))
    except IndexError:
        pass
    try:
        entry["test_keys"]["requests"].append(control_requests.pop(0))
    except IndexError:
        pass
    entry["test_keys"]["requests"] += experiment_requests
    entry["test_keys"]["requests"] += control_requests
    if entry["test_keys"].get("headers_diff", None) is not None:
        entry["test_keys"]["headers_diff"] = list(entry["test_keys"]["headers_diff"])
    return entry


def _normalize_answer(ans):
    try:
        ttl = match(r"ttl=(\d+)", ans[0])
    except Exception:
        log.error("Failed to parse ttl in %s" % ans[0])
        ttl = None

    answer_type = match("type=([A-Z]+)", ans[0])

    na = dict(ttl=ttl, answer_type=answer_type)

    if answer_type == "A":
        na["ipv4"] = match("address=" + regexps["ipv4"], ans[1])

    elif answer_type == "MX":
        na["hostname"] = match("address=" + regexps["ipv4"], ans[1])
        na["preference"] = match(r"preference=(\d+)", ans[1])

    elif answer_type in ["PTR", "CNAME"]:
        na["hostname"] = match("name=" + regexps["hostname"], ans[1])

    elif answer_type == "SOA":
        na["responsible_name"] = match("rname=" + regexps["hostname"], ans[1])
        na["hostname"] = match("mname=" + regexps["hostname"], ans[1])
        na["serial_number"] = match(r"serial=(\d+)", ans[1])
        na["refresh_interval"] = match(r"refresh=(\d+)", ans[1])
        na["retry_interval"] = match(r"retry=(\d+)", ans[1])
        na["minimum_ttl"] = match(r"minimum=(\d+)", ans[1])
        na["expiration_limit"] = match(r"expire=(\d+)", ans[1])

    return na


def normalize_dnst(entry):
    entry["test_keys"].pop("test_resolvers", None)

    errors = entry["test_keys"].pop("tampering", None)
    if errors:
        entry["test_keys"]["errors"] = errors
        entry["test_keys"]["successful"] = [e[0] for e in errors if e[1] is False]
        entry["test_keys"]["failed"] = [e[0] for e in errors if e[1] is not True]
        entry["test_keys"]["inconsistent"] = [e[0] for e in errors if e[1] is True]
    elif entry["test_name"] == "dns_consistency":
        entry["test_keys"]["errors"] = {}
        entry["test_keys"]["successful"] = []
        entry["test_keys"]["failed"] = []
        entry["test_keys"]["inconsistent"] = []

    queries = []
    for query in entry["test_keys"].pop("queries", []):
        try:
            query["hostname"] = match(r"\[Query\('(.+)'", query.pop("query"))
        except:
            query["hostname"] = None

        try:
            query["resolver_hostname"], query["resolver_port"] = query.pop("resolver")
        except:
            query["resolver_hostname"], query["resolver_port"] = [None, None]

        query.pop("addrs", None)

        answers = []
        for answer in query.pop("answers", []):
            normalized_answer = _normalize_answer(answer)
            answers.append(normalized_answer)

        query["answers"] = answers

        failure = query.get("failure", None)
        if not failure and len(answers) == 0:
            failure = "no_answer"
        query["failure"] = failure

        queries.append(query)
    entry["test_keys"]["queries"] = queries
    return entry


def normalize_tcpt(entry):
    return entry


def normalize_process(entry):
    return entry


def normalize_entry(entry, bucket_date, perma_fname, esha):
    """Autoclaving"""
    hashuuid = esha[:16]  # sha1 is 20 bytes

    if isinstance(entry.get("report"), dict):
        entry.update(entry.pop("report"))

    test_name = entry.get("test_name", "invalid")
    test_name = test_name_mappings.get(test_name, test_name.lower())
    entry["test_name"] = test_name

    entry["bucket_date"] = bucket_date

    if not entry.get("id"):
        entry["id"] = str(uuid.UUID(bytes=hashuuid))
    entry["report_filename"] = perma_fname

    # Ensure all the keys in the schema are present
    for key in schema:
        entry[key] = entry.get(key, None)

    if entry.get("data_format_version", "") == "0.2.0":
        if entry["test_keys"] is None:
            entry = nest_test_keys(entry)

        return entry

    ts = entry.pop("start_time", 0)
    test_start_time = datetime.utcfromtimestamp(ts)
    try:
        tst = entry.pop("test_start_time")
        # This is the old test_start_time key that now is called
        # "measurement_start_time"
        if isinstance(tst, float):
            measurement_start_time = datetime.utcfromtimestamp(tst)
        elif tst is None:
            measurement_start_time = test_start_time
        else:
            test_start_time = datetime.strptime(tst, "%Y-%m-%d %H:%M:%S")
            measurement_start_time = datetime.strptime(
                entry.get("measurement_start_time"), "%Y-%m-%d %H:%M:%S"
            )
    except KeyError:
        # Failback to using the start_time
        measurement_start_time = test_start_time

    entry["measurement_start_time"] = measurement_start_time.strftime(
        "%Y-%m-%d %H:%M:%S"
    )
    entry["test_start_time"] = test_start_time.strftime("%Y-%m-%d %H:%M:%S")

    entry["data_format_version"] = "0.2.0"

    if isinstance(entry.get("options", []), dict):
        entry["options"] = entry["options"].get("subargs", [])

    entry = nest_test_keys(entry)

    assert test_name not in test_categories["process"]

    # Some values overlap across categories so multiple ifs are needed
    if test_name in test_categories["httpt"]:
        entry = normalize_httpt(entry)
    if test_name in test_categories["dnst"]:
        entry = normalize_dnst(entry)

    # Ignore old, rare tests
    if test_name in test_categories["scapyt"]:
        raise UnsupportedTestError

    if test_name in ("captive_portal", "tls_handshake"):
        raise UnsupportedTestError

    # TODO: tests these in
    # test_normalize_yaml_sanitise_tcp_connect_bridge_reach
    if entry["test_name"] == "tcp_connect":
        # On 2019-10-08 the last bridge_reachability entry from YAML
        # in the metadb was from 2016-10-12
        raise UnsupportedTestError

    elif entry["test_name"] == "bridge_reachability":
        # On 2019-10-08 the last bridge_reachability entry from YAML
        # in the metadb was from 2016-10-12
        raise UnsupportedTestError

    return entry


### Stream entries from YAML.lz4 files ####


class BlobSlicerError(RuntimeError):
    pass


class BrokenFrameError(BlobSlicerError):
    pass


class TruncatedReportError(BlobSlicerError):
    pass


def stream_yaml_blobs(fd):
    """Detects YAML objects from a stream.
    Returns an iterator of (offset, blob)
    """
    head = b""
    for blob in iter(functools.partial(fd.read, 1048576), ""):
        if len(blob) == 0:
            break
        bloboff = fd.tell() - len(blob)
        head, blob = b"", head + blob
        start = 0
        while head == b"":
            prefix = blob[start : start + 4]
            if prefix == b"---\n":  # ordinary preamble
                end = blob.find(b"\n...\n", start)
                if end != -1:
                    yield bloboff + start, blob[start : end + 5]
                    start = end + 5
                else:
                    head = blob[start:]
            elif not prefix:
                break
            elif prefix == b"...\n":  # duplicate trailer
                # e.g. 2013-05-05/20130505T065614Z-VN-AS24173-dns_consistency-no_report_id-0.1.0-probe.yaml
                start += 4
            elif len(prefix) < 4:  # need next blob
                head = blob[start:]
            elif chr(prefix[0]) == "#":  # comment
                # e.g. 2013-09-12/20130912T144929Z-MD-AS1547-dns_consistency-no_report_id-0.1.0-probe.yaml
                end = blob.find(b"\n", start)
                if end != -1:
                    start = end + 1
                else:
                    head = blob[start:]
            else:
                raise BrokenFrameError(bloboff + start, prefix)

    if head:
        raise TruncatedReportError(fd.tell() - len(head), head[:100])


def generate_report_id(header):
    # TODO: test
    start_time = datetime.fromtimestamp(header.get("start_time", 0))
    report_id = start_time.strftime("%Y%m%dT%H%M%SZ_")
    value_to_hash = header.get("probe_cc", "ZZ").encode("utf-8")
    value_to_hash += header.get("probe_asn", "AS0").encode("utf-8")
    value_to_hash += header.get("test_name", "invalid").encode("utf-8")
    value_to_hash += header.get("software_version", "0.0.0").encode("utf-8")
    probe_city = header.get("probe_city", "None")
    # probe_city = probe_city.encode("utf-8")  # u'Reykjav\xedk' in bucket 2014-02-20
    # probe_city = (
    #   probe_city.encode("utf-8")
    #   if isinstance(probe_city, unicode)
    #   else str(probe_city)
    # )  # u'Reykjav\xedk' in bucket 2014-02-20
    # probe_city = probe_city.decode() #encode("utf-8")
    if probe_city is not None:
        value_to_hash += probe_city.encode("utf-8")
    report_id += "".join(
        string.ascii_letters[b % len(string.ascii_letters)]
        for b in hashlib.sha512(value_to_hash).digest()
    )[:50]
    return report_id


## Entry points


def iter_yaml_msmt_normalized(data, bucket_tstamp: str, report_fn: str):
    """Yields normalized measurements from a YAML bytestream"""
    assert bucket_tstamp.startswith("20")
    assert len(bucket_tstamp) == 10
    assert len(report_fn.split("/")) == 2, report_fn
    # Taken from autoclaving.py stream_yaml_reports
    blobgen = stream_yaml_blobs(data)

    off, header = next(blobgen)
    headsha = hashlib.sha1(header)
    # XXX: bad header kills whole bucket
    header = yaml.safe_load(header)
    if isinstance(header.get("probe_city"), bytes):
        header["probe_city"] = header["probe_city"].decode(errors="ignore")

    # Generates report_id if needed
    if not header.get("report_id"):
        header["report_id"] = generate_report_id(header)

    for off, raw_entry in blobgen:
        esha = headsha.copy()
        esha.update(raw_entry)
        esha_d = esha.digest()
        try:
            entry = yaml.safe_load(raw_entry)
        except yaml.constructor.ConstructorError:
            rid = header["report_id"]
            log.info(f"YAML construction error {rid}")
            entry = None
        except yaml.parser.ParserError:
            rid = header["report_id"]
            log.info(f"YAML parsing error {rid}")
            entry = None

        if not entry:  # e.g. '---\nnull\n...\n'
            continue
        if "test_start_time" in entry and "test_start_time" in header:
            header.pop("test_start_time")
        entry.update(header)
        try:
            d = normalize_entry(entry, bucket_tstamp, report_fn, esha_d)
            msmt_uid = trivial_id(raw_entry, d)
            d["measurement_uid"] = msmt_uid
            yield d
        except Exception as e:
            log.error(f"Skipping measurement: {e}")
            continue



try:
    import importlib.metadata as importlib_metadata
except ModuleNotFoundError:
    import importlib_metadata  # type: ignore it's only for py <=3.8

__version__ = importlib_metadata.version(__name__)



from typing import Optional
import orjson
import requests


def get_raw_measurement(report_id: str, input: Optional[str] = None) -> bytes:
    params = params = {"report_id": report_id, "full": True}
    if input:
        params["input"] = input
    r = requests.get("https://api.ooni.io/api/v1/measurement_meta", params=params)
    j = r.json()
    return j["raw_measurement"].encode("utf-8")


def get_measurement_dict(report_id: str, input: Optional[str] = None) -> dict:
    params = params = {"report_id": report_id, "full": True}
    if input:
        params["input"] = input
    r = requests.get("https://api.ooni.io/api/v1/measurement_meta", params=params)
    j = r.json()
    msmt_dict = orjson.loads(j["raw_measurement"])
    msmt_dict["measurement_uid"] = j["measurement_uid"]
    return msmt_dict



import logging
from typing import List, Optional, NamedTuple
from enum import Enum
from datetime import datetime
from dataclasses import dataclass


from oonidata.observations import (
    Observation,
)

log = logging.getLogger("oonidata.events")


class BlockingType(Enum):
    # k: everything is OK
    OK = "k"
    # b: blocking is happening with an unknown scope
    BLOCKED = "b"
    # n: national level blocking
    NATIONAL_BLOCK = "n"
    # i: isp level blocking
    ISP_BLOCK = "i"
    # l: local blocking (school, office, home network)
    LOCAL_BLOCK = "l"
    # s: server-side blocking
    SERVER_SIDE_BLOCK = "s"
    # d: the subject is down
    DOWN = "d"
    # t: this is a signal indicating some form of network throttling
    THROTTLING = "t"


def fp_scope_to_outcome(scope: Optional[str]) -> BlockingType:
    # "nat" national level blockpage
    # "isp" ISP level blockpage
    # "prod" text pattern related to a middlebox product
    # "inst" text pattern related to a voluntary instition blockpage (school, office)
    # "vbw" vague blocking word
    # "fp" fingerprint for false positives
    if scope == "nat":
        return BlockingType.NATIONAL_BLOCK
    elif scope == "isp":
        return BlockingType.ISP_BLOCK
    elif scope == "inst":
        return BlockingType.LOCAL_BLOCK
    elif scope == "fp":
        return BlockingType.SERVER_SIDE_BLOCK
    return BlockingType.BLOCKED


class BlockingEvent(NamedTuple):
    blocking_type: BlockingType
    blocking_subject: str
    blocking_detail: str
    blocking_meta: dict
    confidence: float


@dataclass
class ExperimentResult:
    measurement_uid: str
    report_id: str
    input: str
    timestamp: datetime

    probe_asn: int
    probe_cc: str

    probe_as_org_name: str
    probe_as_cc: str

    network_type: str

    resolver_ip: Optional[str]
    resolver_asn: Optional[int]
    resolver_as_org_name: Optional[str]
    resolver_as_cc: Optional[str]
    resolver_cc: Optional[str]

    observation_ids: List[str]
    blocking_events: List[BlockingEvent]
    ok_confidence: float

    anomaly: bool
    confirmed: bool


def make_base_result_meta(obs: Observation) -> dict:
    return dict(
        measurement_uid=obs.measurement_uid,
        report_id=obs.report_id,
        input=obs.input,
        timestamp=obs.timestamp,
        probe_asn=obs.probe_asn,
        probe_cc=obs.probe_cc,
        probe_as_org_name=obs.probe_as_org_name,
        probe_as_cc=obs.probe_as_cc,
        network_type=obs.network_type,
        resolver_ip=obs.resolver_ip,
        resolver_asn=obs.resolver_asn,
        resolver_as_org_name=obs.resolver_as_org_name,
        resolver_as_cc=obs.resolver_as_cc,
        resolver_cc=obs.resolver_cc,
    )
